"""
Calculations of CASTEP
"""
from __future__ import print_function
import warnings

import aiida
from aiida.common.exceptions import InputValidationError
from aiida.common.utils import classproperty
from aiida.orm import CalculationFactory, DataFactory
from aiida_castep.calculations.base import BaseCastepInputGenerator
from aiida_castep.calculations.base import __version__ as base_version

from .utils import get_castep_ion_line

from .._version import calc_parser_version
__version__ = calc_parser_version


if aiida.__version__.startswith("0"):
    JobCalculation = CalculationFactory("job", True)
else:
    JobCalculation = CalculationFactory("job")

KpointsData = DataFactory("array.kpoints")
StructureData = DataFactory("structure")
ParameterData = DataFactory("parameter")

# Define the version of the calculation

class CastepCalculation(BaseCastepInputGenerator, JobCalculation):
    """
    Class representing a generic CASTEP calculation -
    This class should work for all types of calculations.
    """

    _default_symlink_usage = True

    # NOT CURRENTLY USED
    _acceptable_tasks = [
        "singlepoint",
        "geometryoptimization",
        "geometryoptimisation",
    ]

    _copied_attributes = ["jobresource_param", 
                          "custom_scheduler_commands", 
                          "max_wallclock_seconds"]

    def _init_internal_params(self):

        super(CastepCalculation, self)._init_internal_params()
        self._default_parser = "castep.castep"
        self._use_kpoints = True
        self._SEED_NAME = "aiida"
        self._DEFAULT_INPUT_FILE = "aiida.cell"
        self._DEFAULT_OUTPUT_FILE = "aiida.castep"

    @classproperty
    def _use_methods(cls):
        """
        Extend the parent _use_methods with further keys
        """

        retdict = JobCalculation._use_methods
        retdict.update(BaseCastepInputGenerator._baseclass_use_methods)

        # Not all calculation need kpoints
        retdict['kpoints'] = {
            'valid_types': KpointsData,
            'additional_parameter': None,
            'linkname': 'kpoints',
            'docstring': "Use the node defining the kpoint sampling to use",
        }

        return retdict

    def submit_test(self,
                    dryrun=False,
                    restart_check=False,
                    verbose=True,
                    castep_exe="castep.serial",
                    **kwargs):
        """
        Test submission. Optionally do a local dryrun.
        Return and dictionary as the third item
        * num_kpoints: number of kpoints used.
        * memory_MB: memory usage estimated MB
        * disk_MB: disk space usage estimated in MB
        """
        outcome = super(CastepCalculation, self).submit_test(**kwargs)
        outinfo = {}
        outinfo["_submit_test"] = outcome
        if dryrun:
            folder, dryrun_results = self._dryrun_test(outcome[0], castep_exe,
                                                       verbose)
            outinfo["dryrun_results"] = dryrun_results

        if restart_check:
            self.check_restart(verbose)
        return outcome

    def _check_restart(self, verbose=True):
        """
        Check the existence of restart file if requested
        """
        import os
        from .utils import _lowercase_dict

        def _print(inp):
            if verbose:
                print(inp)

        inps = self.get_inputs_dict()
        paramdict = inps[self.get_linkname("parameters")].get_dict()["PARAM"]

        paramdict = _lowercase_dict(paramdict, "paramdict")
        stemp = paramdict.get("reuse", None)
        if not stemp:
            stemp = paramdict.get("continuation", None)

        if stemp is not None:
            fname = os.path.split(stemp)[-1]
        else:
            # No restart file needed
            _print("This calculation does not require a restart file.")
            return

        # Now check if the remote folder has this file
        remote_data = inps.get(self.get_linkname("parent_folder"), None)
        if not remote_data:
            raise InputValidationError(
                "Restart requires "
                "partent_folder to be specified".format(fname))
        else:
            folder_list = remote_data.listdir()
            if fname not in folder_list:
                raise InputValidationError(
                    "Restart file {}"
                    " is not in the remote folder".format(fname))
            else:
                _print("Check finished, restart file '{}' exists.".format(fname))

    def _dryrun_test(self, folder, castep_exe, verbose=True):
        """
        Do a dryrun test in a folder with prepared inputs
        """

        from fnmatch import fnmatch

        def _print(inp):
            if verbose:
                print(inp)

        # Do a dryrun
        from subprocess import call, check_output
        try:
            output = check_output([castep_exe, "-v"]).decode()
        except OSError:
            _print("CASTEP executable '{}' is not found".format(castep_exe))
            return

        # Now start dryrun
        _print("Running with {}".format(
            check_output(["which", castep_exe]).decode()))
        _print(output)

        _print("Starting dryrun...")
        call([castep_exe, "--dryrun", self._SEED_NAME], cwd=folder.abspath)

        # Check if any *err files
        contents = folder.get_content_list()
        for n in contents:
            if fnmatch(n, "*.err"):
                with folder.open(n) as fh:
                    _print("Error found in {}:\n".format(n))
                    _print(fh.read())
                raise InputValidationError("Error found during dryrun")

        # Gather information from the dryrun file
        import re
        dryrun_results = {}
        with folder.open(self._DEFAULT_OUTPUT_FILE) as fh:
            for line in fh:
                mth = re.match(r"\s*k-Points For SCF Sampling:\s+(\d+)\s*",
                               line)
                if mth:
                    dryrun_results["num_kpoints"] = int(mth.group(1))
                    _print("Number of k-points: {}".format(mth.group(1)))
                    mth = None
                    continue
                mth = re.match(
                    r"\| Approx\. total storage required"
                    r" per process\s+([0-9.]+)\sMB\s+([0-9.]+)", line)
                if mth:
                    dryrun_results["memory_MB"] = (float(mth.group(1)))
                    dryrun_results["disk_MB"] = (float(mth.group(2)))
                    _print("RAM: {} MB, DISK: {} MB".format(
                        mth.group(1), mth.group(2)))
                    mth = None
                    continue

        return folder, dryrun_results


    def duplicate(self):
        """
        Duplicate this calculation return an new, unstore calculation with
        the same attributes but no links attached. label and descriptions
        are also copied.
        """
        new = type(self)()

        attrs = self.get_attrs()
        if attrs:
            for k, v in attrs.items():
                if k not in self._updatable_attributes:
                    new._set_attr(k, v)

        new.label = self.label
        new.description = self.description
        # Set the computer as well
        new.set_computer(self.get_computer())

        return new

    def compare_with(self, the_other_calc, reverse=False):
        """
        Compare with another calculation
        Look for difference in get_castep_input_summary functions
        :params node: pk or uuid or node
        :params reverse: reverse the comparison, by default this node
        is the "new" and the one compared with is "old".
        """
        if isinstance(the_other_calc, (int, basestring)):
            from aiida.orm import load_node
            calc2 = load_node(the_other_calc)
        else:
            calc2 = the_other_calc

        from deepdiff import DeepDiff
        this_param = self.get_castep_input_summary()
        other_param = calc2.get_castep_input_summary()
        if reverse is True:
            res = DeepDiff(this_param, other_param)
        else:
            res = DeepDiff(other_param, this_param)

        # Compare the kpoints


#        this_kpt = self.get_inputs_dict()[self.get_linkname('kpoints')]
#        other_kpy = calc2.get_inputs_dict()[calc2.get_linkname('kpoints')]

# Compare psudo
        return res

    def get_castep_input_summary(self):
        """
        Convenient fuction for getting a summary of the
        input of this calculation
        """

        from aiida_castep.data import get_pseudos_from_structure
        UpfData = DataFactory("upf")
        OTFGData = DataFactory("castep.otfgdata")
        UspData = DataFactory("castep.uspdata")

        out_info = {}

        inp_dict = self.get_inputs_dict()
        in_param = inp_dict[self.get_linkname('parameters')]
        in_kpn = inp_dict[self.get_linkname('kpoints')]
        in_settings = inp_dict.get(self.get_linkname('settings'), None)
        in_structure = inp_dict[self.get_linkname('structure')]
        in_code = inp_dict[self.get_linkname('code')]

        # Snippet from Pseudos calculation
        pseudos = {}
        # A dictionary that associates each kind name to a pseudo
        for link in inp_dict.keys():
            if link.startswith(self._get_linkname_pseudo_prefix()):
                kindstring = link[len(self._get_linkname_pseudo_prefix()):]
                kinds = kindstring.split('_')
                the_pseudo = inp_dict.pop(link)
                if not isinstance(the_pseudo, (UpfData, UspData, OTFGData)):
                    raise InputValidationError(
                        "Pseudo for kind(s) {} is not of "
                        "supoorted ".format(",".join(kinds)))
                for kind in kinds:
                    if kind in pseudos:
                        raise InputValidationError(
                            "Pseudo for kind {} passed "
                            "more than one time".format(kind))
                    if isinstance(the_pseudo, OTFGData):
                        pseudos[kind] = the_pseudo.string
                    elif isinstance(the_pseudo, (UspData, UpfData)):
                        pseudos[kind] = the_pseudo.filename


        param_dict = in_param.get_dict()
        out_info.update(param_dict)

        out_info["kpoints"] = in_kpn.get_desc()
        out_info["structure"] = {
            "formula": in_structure.get_formula(),
            "cell": in_structure.cell,
            "label": in_structure.label
        }

        out_info["code"] = in_code
        out_info["computer"] = self.get_computer()
        out_info["resources"] = self.get_resources()
        out_info["custom_scheduler_commands"] = self.get_custom_scheduler_commands()
        out_info["wallclock"] = self.get_max_wallclock_seconds()

        if in_settings is not None:
            out_info["settings"] = in_settings.get_dict()
        out_info["label"] = self.label
        out_info["pseudos"] = pseudos
        return out_info


    def update_parameters(self, force=False, delete=None, **kwargs):
        """
        Convenient function to update the parameters of the calculation.
        Will atomiatically set the PARAM or CELL field in unstored
        ParaemterData linked to the calculation.
        If no ``ParameterData`` is linked to the calculation, a new node will be
        created.

        ..note:
          This method relies on the help information to check and assign
          keywords to PARAM or CELL field of the ParameterData
          (i.e for generating .param and .cell file)


        calc.update_parameters(task="singlepoint")

        :param force: flag to force the update even if the ParameterData node is stored.
        :param delete: A list of the keywords to be deleted.
        """
        param_node = self.get_inputs_dict().get(self.get_linkname('parameters'), None)
        # Create the node if none is found
        if param_node is None:
            warnings.warn("No existing ParameterData node found, creating a new one.")
            param_node = ParameterData(dict={"CELL": {}, "PARAM": {}})
            self.use_parameters(param_node)

        if param_node.is_stored:
           if force:
               # Create a new node if the existing node is stored
               param_node = ParameterData(dict=param_node.get_dict())
           else:
            raise RuntimeError("The input ParameterData<{}> is already stored".format(param_node.pk))

        param_dict = param_node.get_dict()

        # Update the dictionary
        from .helper import HelperCheckError
        helper = self.get_input_helper()
        dict_update, not_found = helper._from_flat_dict(kwargs)
        if not_found:
            suggest = [helper.get_suggestion(i) for i in not_found]
            error_string = "Following keys are invalid -- "
            for error_key, sug in zip(not_found, suggest):
                error_string += "{}: {}; ".format(error_key, sug)
            raise HelperCheckError(error_string)
        else:
            param_dict["PARAM"].update(dict_update["PARAM"])
            param_dict["CELL"].update(dict_update["CELL"])

        # Delete any keys as requested
        if delete:
            for key in delete:
                tmp1 = param_dict["PARAM"].pop(key, None)
                tmp2 = param_dict["CELL"].pop(key, None)
                if (tmp1 is None) and (tmp2 is None):
                    raise RuntimeError("Key {} not found".format(key))

        # Apply the change to the node
        param_node.set_dict(param_dict)


class Pot1dCalculation(CastepCalculation):
    """
    Class for pot1d Calculation
    """

    _internal_retrieve_list = CastepCalculation.\
                              _internal_retrieve_list + ["*.dat"]

    def _init_internal_params(self):
        super(Pot1dCalculation, self)._init_internal_params()
        self._default_parser = "castep.pot1d"

    @classmethod
    def from_calculation(cls, calc, code, use_castep_bin=False, **kwargs):
        """
        Create pot1d calculation using existing calculation.
        ``code`` must be specified as it is different from the original CASTEP code.
        """

        out_calc = cls.continue_from(
            calc,
            ignore_state=True,
            restart_type="continuation",
            use_output_structure=True,
            use_castep_bin=use_castep_bin,
            **kwargs)
        out_calc.use_code(code)
        return out_calc

    def _generate_CASTEPinputdata(self, *args, **kwargs):
        out = super(Pot1dCalculation, self).\
              _generate_CASTEPinputdata(*args, **kwargs)
        if out[1].get("continuation") is None:
            raise InputValidationError("pot1d requires "
                                       "continuation being set in .param")
        return out

    def get_withmpi(self):
        """
        pot1d is not compile with mpi.
        Hence the default is changed to False instead.
        """
        return self.get_attr('withmpi', False)


class TaskSpecificCalculation(CastepCalculation):
    """
    Class for Calculations that only allow certain tasks
    """

    _acceptable_tasks = []

    def _generate_CASTEPinputdata(self, *args, **kwargs):
        param = args[0].get_dict()

        # Check if task is correctly set
        if param['PARAM']['task'].lower() not in map(str.lower,
                                                     self._acceptable_tasks):
            raise InputValidationError("Wrong TASK value {}"
                                       " set in PARAM".format(
                                           param['PARAM']['task'].lower()))
        return super(TaskSpecificCalculation, self)._generate_CASTEPinputdata(
            *args, **kwargs)


class CastepTSCalculation(TaskSpecificCalculation):
    """
    CASTEP calculation for transition state search. Use an extra input product structure.
    """
    _acceptable_tasks = ["transitionstatesearch"]

    @classproperty
    def _use_methods(cls):

        retdict = super(CastepTSCalculation, cls)._use_methods
        retdict['product_structure'] = {
            'valid_types':
            StructureData,
            'additional_parameter':
            None,
            'linkname':
            'product_structure',
            'docstring':
            "Use the node defining the structure as the product structure in transition state search."
        }
        return retdict

    def _generate_CASTEPinputdata(self, *args, **kwargs):
        """
        Override superclass methods
        """
        cell, param, local_copy = super(CastepTSCalculation, self).\
                    _generate_CASTEPinputdata(*args, **kwargs)
        p_structure = kwargs[self.get_linkname('product_structure')]

        pdt_position_list = []
        for site in p_structure.sites:
            kind = p_structure.get_kind(site.kind_name)
            name = kind.symbol
            line = get_castep_ion_line(name, site.position)
            pdt_position_list.append(line)

        cell["POSITIONS_ABS_PRODUCT"] = pdt_position_list
        return cell, param, local_copy


class CastepExtraKpnCalculation(TaskSpecificCalculation):
    """
    CASTEP calculation with extra kpoints (e.g SPEC, BS, PHONON, SPECTRAL)
    """
    KPN_NAME = ""  # Alias of the name, e.g BS for bandstructure calculation
    CHECK_EXTRA_KPN = False  # Check the existence of extra kpoints node

    @classproperty
    def kpn_name(cls):
        return cls.KPN_NAME.lower()

    @classproperty
    def _use_methods(cls):

        retdict = CastepCalculation._use_methods

        retdict['{}_kpoints'.format(cls.KPN_NAME.lower())] = {
            'valid_types':
            KpointsData,
            'additional_parameter':
            None,
            'linkname':
            '{}_kpoints'.format(cls.kpn_name),
            'docstring':
            "Use the node defining the kpoint sampling for band {}  calculation"
            .format(cls.TASK.lower())
        }
        return retdict

    def _generate_CASTEPinputdata(self, *args, **kwargs):
        """Add BS kpoints information to the calculation"""

        cell, param, local_copy = super(CastepExtraKpnCalculation,
                                        self)._generate_CASTEPinputdata(
                                            *args, **kwargs)

        # Check the existence of extra kpoints
        try:
            extra_kpns = kwargs[self.get_linkname('{}_kpoints'.format(
                self.kpn_name))]
        except KeyError:
            if self.CHECK_EXTRA_KPN:
                raise InputValidationError("{}_kpoints"
                                           " node not found".format(
                                               self.kpn_name))
            else:
                return cell, param, local_copy

        # Add information in the node to cell file
        try:
            mesh, offset = extra_kpns.get_kpoints_mesh()
            has_mesh = True
        except AttributeError:
            try:
                bs_kpts_list = extra_kpns.get_kpoints()
                num_kpoints = len(bs_kpts_list)
                has_mesh = False
                if num_kpoints == 0:
                    raise InputValidationError(
                        "At least one k points must be provided")
            except AttributeError:
                raise InputValidationError(
                    "No valid {}_kpoints have been found".format(
                        self.kpn_name))

            try:
                _, weights = extra_kpns.get_kpoints(also_weights=True)

            except AttributeError:
                import numpy as np
                weights = np.ones(num_kpoints, dtype=float) / num_kpoints

        if has_mesh is True:
            mesh_name = "{}_kpoints_mp_grid".format(self.kpn_name)
            cell[mesh_name] = "{} {} {}".format(*mesh)
            if offset != [0., 0., 0.]:
                cell[mesh_name.replace("grid",
                                       "offset")] = "{} {} {}".format(*offset)
        else:
            bs_kpts_lines = []
            for kpoint, weight in zip(bs_kpts_list, weights):
                bs_kpts_lines.append("{:18.10f} {:18.10f} "
                                     "{:18.10f} {:18.10f}".format(
                                         kpoint[0], kpoint[1], kpoint[2],
                                         weight))
            bname = "{}_kpoints_list".format(self.kpn_name).upper()
            cell[bname] = bs_kpts_lines
        return cell, param, local_copy

    def create_restart(self, *args, **kwargs):
        """
        Create a restart of the calculation
        """
        out_calc = super(CastepExtraKpnCalculation, self).create_restart(
            *args, **kwargs)

        # Attach the extra kpoints node if it is there
        inp_name = "{}_kpoints".format(self.kpn_name)  # Name of the input
        linkname = self.get_linkname(inp_name)  # Name of the link

        extra_kpn_node = self.get_inputs_dict().get(linkname)
        if extra_kpn_node:
            getattr(out_calc, "use_" + inp_name)(extra_kpn_node)
        return out_calc

    def get_castep_input_summary(self):
        """
        Generate a dictionary to summarize the inputs to CASTEP
        """

        inp_name = "{}_kpoints".format(self.kpn_name)  # Name of the input
        linkname = self.get_linkname(inp_name)  # Name of the link

        out_dict = super(CastepExtraKpnCalculation, self).get_castep_input_summary()
        out_dict[inp_name] = self.get_inputs_dict().get(linkname)
        return out_dict

    @classmethod
    def continue_from(cls, *args, **kwargs):
        """
        Create a new calculation as a continuation from a given calculation.
        This is effectively an "restart" for CASTEP and a lot of the parameters
        can be tweaked. For example, conducting bandstructure calculation from
        finished geometry optimisation's.
        :param bool ignore_state: Ignore the state of parent calculation
        :param str restart_type: "continuation" or "restart".
        If set to continuation the child calculation has keyword
        'continuation' set.
        :param bool reuse: Whether we want to reuse the previous calculation.
        only applies for "restart" run
        :param bool parent_folder_symlink: if True, symlink are used instead
        of hard copies of the files. Default given be
        self._default_symlink_usage
        :param bool use_output_structure: if True, the output structure of
        parent calculation is used as the input of the child calculation.
        This is useful for photon/bs calculation.

        See also: create_restart
        """
        cout = super(CastepExtraKpnCalculation, cls).continue_from(
            *args, **kwargs)

        # Check the task keyword
        param = cout.get_inputs_dict()[cout.get_linkname('parameters')]
        param_dict = param.get_dict()

        task = param_dict['PARAM'].get('task')
        if task and task == cls.TASK.lower():
            pass
        else:
            # Replace task
            param_dict['PARAM']['task'] = cls.TASK.lower()
            from aiida.orm import DataFactory
            ParameterData = DataFactory('parameter')
            new_param = ParameterData(dict=param_dict)
            cout._remove_link_from(cout.get_linkname('parameters'))
            cout.use_parameters(new_param)

        return cout


class CastepBSCalculation(CastepExtraKpnCalculation):
    """
    CASTEP bandstructure calculation
    """

    TASK = "BANDSTRUCTURE"
    _acceptable_tasks = [TASK]
    KPN_NAME = "BS"


class CastepSpectralCalculation(CastepExtraKpnCalculation):
    """
    CASTEP spectral calculation
    """
    TASK = "SPECTRAL"
    _acceptable_tasks = [TASK]
    KPN_NAME = "SPECTRAL"


class CastepOpticsCalclulation(CastepExtraKpnCalculation):
    """
    CASTEP Optics calculation
    """
    TASK = "OPTICS"
    _acceptable_tasks = [TASK]
    KPN_NAME = "OPTICS"



"""
WorkChain for doing an relaxation, try to handle erros including:
run not finished
geom not converged within given cycle
scf iteration reached maximum

and possible adapt from it
"""
from copy import deepcopy
import time

from aiida.orm import Code
from aiida.orm.data.base import Str, Float, Bool, Int
from aiida.orm.data.folder import FolderData
from aiida.orm.data.remote import RemoteData
from aiida.orm.data.parameter import ParameterData
from aiida.orm.data.structure import StructureData
from aiida.orm.data.array.kpoints import KpointsData
from aiida.orm.data.singlefile import SinglefileData
from aiida.orm.group import Group
from aiida.orm.utils import WorkflowFactory
from aiida.common.exceptions import AiidaException, NotExistent
from aiida.common.datastructures import calc_states
from aiida.work.run import submit
from aiida.work.workchain import WorkChain, if_, while_, append_
from aiida.orm.utils import CalculationFactory

from aiida_castep.parsers.raw_parser import (END_NOT_FOUND_MESSAGE,
 GEOM_FAILURE_MESSAGE, SCF_FAILURE_MESSAGE)
from aiida_castep.data import get_pseudos_from_structure

CastepCalculation = CalculationFactory('castep.castep')


class UnexpectedFailure(RuntimeError):
    pass

class CastepRelaxWorkChain(WorkChain):
    """
    Workchain to do geometryoptimisation with CASTEP
    Try to restart calculations unti convergence is reached
    """
    SUBMISSION_RETRY_SECONDS = 60
    MAX_ITERATIONS = 5

    @classmethod
    def define(cls, spec):
        super(CastepRelaxWorkChain, cls).define(spec)
        spec.input('code', valid_type=Code)
        spec.input('structure', valid_type=StructureData)
        spec.input('parameters', valid_type=ParameterData)
        spec.input('settings', valid_type=ParameterData, required=False)
        spec.input('pseudo_family', valid_type=Str, required=False)
        spec.input_group('pseudos', valid_type=Str, required=False)
        spec.input('kpoints', valid_type=KpointsData)
        spec.input('options', valid_type=ParameterData, required=False)
        spec.input('final_restart', valid_type=Bool, required=False, default=Bool(False))
        spec.input('geom_method', valid_type=Str, default=Str('LBFGS'))
        spec.outline(
            cls.initialise,
            cls.validate_inputs,
            while_(cls.should_run_relax)(
                cls.run_relax,
                cls.check_relax,
            ),
            if_(cls.should_do_final_restart)(
                cls.run_final_restart,
            ),
            cls.process_results,
            )
        spec.output('output_structure', valid_type=StructureData)
        spec.output('output_parameters', valid_type=ParameterData)
        spec.output('remote_folder', valid_type=RemoteData)
        spec.output('retrieved', valid_type=FolderData)

    def initialise(self):
        """
        Initialise the workchain. Process inputs, and save it to ctx
        """
        self.ctx.current_parent_folder = None
        self.ctx.is_converged = False
        self.ctx.iteration = 0
        self.ctx.max_iterations = self.MAX_ITERATIONS
        self.ctx.restart_calc = None
        self.ctx.submission_failure = None

        self.ctx.inputs = {
            'code': self.inputs.code,
            'structure': self.inputs.structure,
            'parameters': self.inputs.parameters.get_dict(),
            'kpoints': self.inputs.kpoints,
            '_options': self.inputs.options.get_dict()
        }

        if "settings" in self.inputs:
            self.ctx.inputs.update(settings=self.inputs.settings)

        # Check if a pseudo family has been passed

        if self.inputs.pseudo_family:
            self.ctx.inputs['pseudo'] = get_pseudos_from_structure(
                self.inputs.structure, self.inputs.pseudo_family.value
                )

        elif self.inputs.pseudos:
            self.ctx.inputs['pseudo'] = self.inputs.pseudos
        else:
            self.abort_nowait('No valida pseudo data passed. Both pseudo_family and pseudos are not specified')
            return

        # Check if task:geometryoptimisation is passed

        # OK this should not happen but I fix them nevertheless here
        if 'PARAM' not in self.ctx.inputs['parameters']:
            self.ctx.inputs['PARAM'] = {}
        if 'CELL' not in self.ctx.inputs['parameters']:
            self.ctx.inputs['CELL'] = {}

        param_dict = self.ctx.inputs['parameters']['PARAM']
        if 'task' not in param_dict:
            param_dict['task'] = 'geometryoptimisation'
        elif param_dict['task'] != 'geometryoptimisation' \
                and param_dict['task'] != 'geometryoptimization':
            self.abort_nowait('Wrong task value passed {}'.format(param_dict['task']))
            return


        # Add correct geom method
        geom_method = param_dict.get('geom_method', None)
        if geom_method is None:
            param_dict['geom_method'] = self.inputs.geom_method.value
        elif geom_method != self.inputs.geom_method:
            self.abort_nowait('Inconsistent geom_method passed. In PARAM it is {} but the input is {}'.format(geom_method, self.inputs.geom_method))
            return

        return

    def validate_inputs(self):
        """
        A more detailed check
        """
        from aiida_castep.calculations.helper import CastepHelper
        helper = CastepHelper()
        # Automatic fix
        out = helper.check_dict(self.ctx.inputs['parameters'], auto_fix=True)
        if out != self.ctx.inputs['parameters']:
            self.report("Auto fixed input parameters. New parameters {}".format(out))

        self.ctx.inputs['parameters'] = out

        # Check if write_checkpoint is disabled
        check_point = out['PARAM'].get("write_checkpoint")
        if check_point:
            if check_point == "none":
                self.report("Warning: checkpointing disabled!")

        opt_strategy  = out['PARAM'].get('opt_strategy')

        if not opt_strategy:
            out['PARAM']['opt_strategy'] = "speed"
            self.report("Automatically set opt_strategy : speed.")

        # Converge back into ParameterData
        # Make a new ParameterData node for calculation

        return

    def should_run_relax(self):
        return not self.ctx.is_converged

    def should_run_final_restart(self):
        return self.inputs.final_restart

    def run_relax(self):
        """
        Run a CastepCalculation
        """
        self.ctx.iteration += 1

        inputs = deepcopy(self.ctx.inputs)

        if self.ctx.restart_calc:
            # We need to restart from a provious calculation
            inputs['parameters']['PARAM']['continuation'] = \
                CastepCalculation.get_restart_file_relative_path(
                    self.ctx.inputs['parameters'], False)
            inputs['parent_folder'] = self.ctx.restart_calc.out.remote_folder

        # Finall make a ParameterData from the inputs
        inputs["parameters"] = ParameterData(dict=inputs["parameters"])
        process = CastepCalculation.process()
        calc = submit(process, **inputs)

        self.report('Launching CastepCalculation <{}> iteration #{}'.format(calc.pid, self.ctx.iteration))


        return self.to_context(calcs=append_(calc))

    def check_relax(self):
        """
        Check ouput of previous relaxation.
        """

        try:
            calc = self.ctx.calcs[-1]

        except IndexError:
            self.abort_nowait('Seems the initial iteration had problems. No calculation had been returned')
            return

        # Possible states that we can fix
        acceptable_states = [calc_states.FINISHED, calc_states.FAILED,
                             calc_states.SUBMISSIONFAILED ]

        # If the calculation is finished
        # What the difference between using get_state() == FINISHED?
        calc_state = calc.get_state()
        if calc.has_finished_ok():
            self.report('Geometry optimisation succesful after {} WorkChain iterations'.format(self.ctx.iteration))
            self.ctx.is_converged = True

        elif self.ctx.iteration >= self.ctx.max_iterations:
            self.report('Has reached max iterations {}'.format(
                self.ctx.max_iterations))
            self.abort_nowait('Last ran CastepCalculation<{}>'.format(calc.pk))
            return

        elif calc_state not in acceptable_states:
            self.abort_nowait('Last ran CastepCalculation <{}> state {} cannot be handled'.format(calc.pk, calc.get_state()))
            return

        elif calc_state in  [calc_states.SUBMISSIONFAILED]:
            self._submission_failure_handler(calc)
            self.ctx.submission_failure = True

        elif calc_state in [calc_states.FAILED]:
            self.ctx.submission_failure = False

            try:
                handler_out = self._failure_handler(calc)
            except UnexpectedFailure as e:
                self.abort_nowait(e.args)
                return

            if handler_out == "CAN RESTART":
                # Use the last calculation to restart
                self.ctx.restart_calc = calc
                return
            else:
                self.abort_nowait("Abort due to failure at iteration {} with CastepCalculation <{}> Message: {}".format(self.ctx.iteration, calc.pk, handler_out))
                return

    def should_do_final_restart(self):
        return False

    def run_final_restart(self):
        pass

    def process_results(self):
        """
        Process the outputs
        """
        last_calc = self.ctx.calcs[-1]
        output_dict = last_calc.get_outputs_dict()
        self.out('output_structure', output_dict["output_structure"])
        self.out('output_parameters', output_dict["output_parameters"])
        self.out('remote_folder', output_dict["remote_folder"])
        self.out('retrieved', output_dict["retrieved"])

    def _submission_failure_handler(self, calc):
        """We try again in 60 seconds"""
        if self.ctx.submission_failure:
            self.abort_nowait('Submission for CastepCalculation <{}> failed twice'.format(calc.pk))
            return
        else:
            self.report('Submission for CastepCalculation <{}> failed. Sleep for {} seconds and try again'.format(calc.pk, self.SUBMISSION_RETRY_SECONDS))
            time.sleep(self.SUBMISSION_RETRY_SECONDS)

    def _failure_handler(self, calc):
        """
        Handle a failed calculation
        """
        try:
            output_parameters = calc.get_outputs_dict()["output_parameters"].get_dict()
        except KeyError:
            raise UnexpectedFailure("No ouput ParameterData find")
        try:
            warnings = output_parameters["warnings"]
        except KeyError:
            raise UnexpectedFailure("Cannot find warnings")

        if any([SCF_FAILURE_MESSAGE in w for w in warnings]):

            return "SCF CONVERGE FAILURE"

        if any([GEOM_FAILURE_MESSAGE in w for w in warnings]):
            self.report("Run finished but geom convergence not reached")
            return "CAN RESTART"

        # Calculation gets killed
        if  any([END_NOT_FOUND_MESSAGE in w for w in warnings]):
            try:
                output_trajectory = calc.get_outputs_dict()["output_trajectory"]
                total_energy = output_trajectory.get_array("total_energy")
            except KeyError:
                self.report("First SCF not finished")
                return "FIRST SCF UNFINISHED"
            else:
                # Check if we have valid energies
                if len(total_energy) > 0:
                    self.report("Run did not finished after {} iterations".format(len(total_energy)))
                    return "CAN RESTART"
                else:
                    return "FIRST SCF UNFINISHED"
        else:
            return "CANNOT FIND VALID WARNINGS"


class TwoStepRelax(WorkChain):
    """
    Two step relaxation. First relax with variable cell, then do a fixed cell
    geometry optimisation. Even if the calcualtion has not convered
    """

    @classmethod
    def define(cls, spec):
        super(TwoStepRelax, cls).define(spec)
        spec.input('code', valid_type=Code)
        spec.input('structure', valid_type=StructureData)
        spec.input('parameters', valid_type=ParameterData)
        spec.input('settings', valid_type=ParameterData, required=False)
        spec.input('pseudo_family', valid_type=Str, required=False)
        spec.input_group('pseudos', valid_type=Str, required=False)
        spec.input('kpoints', valid_type=KpointsData)
        spec.input('options', valid_type=ParameterData, required=False)
        spec.input('var_cell_geom_method', valid_type=Str, default=Str('TPSD'))
        spec.input('var_cell_geom_iter', valid_type=Int, default=Int(20))
        spec.input('fix_cell_geom_method', valid_type=Str, default=Str('TPSD'))
        spec.outline(
            cls.validate_inputs,
            cls.relax,
            cls.swap_param,
            cls.relax,
            cls.process_results,
            )
        spec.output('output_structure', valid_type=StructureData)
        spec.output('output_parameters', valid_type=ParameterData)
        spec.output('remote_folder', valid_type=RemoteData)
        spec.output('retrieved', valid_type=FolderData)

    def validate_inputs(self):

        # ctx.inputs is a template to for castepcalculation processes
        options = self.inputs.options.get_dict()
        self.ctx.inputs = {
            'code': self.inputs.code,
            'structure': self.inputs.structure,
            'kpoints': self.inputs.kpoints,
            '_options': options,
            '_label': options.pop("_calc_label", None),
            '_description': options.pop("_calc_description", None),
        }
        if "settings" in self.inputs:
            self.ctx.inputs["settings"] = self.inputs.settings

        # Setup pseudo potentials
        if self.inputs.pseudo_family:
            self.ctx.inputs['pseudo'] = get_pseudos_from_structure(
                self.inputs.structure, self.inputs.pseudo_family.value
                )
        elif self.inputs.pseudos:
            self.ctx.inputs['pseudo'] = self.inputs.pseudos
        else:
            self.abort_nowait('No valida pseudo data passed. Both pseudo_family and pseudos are not specified')
            return

        # Check possible mistake in the parameterse dictionary
        v_param_dict = self.inputs.parameters.get_dict()
        v_param_dict["CELL"].pop("fix_all_cell", None)
        v_param_dict["PARAM"]["geom_method"] = str(self.inputs.var_cell_geom_method)
        v_param_dict["PARAM"]["geom_max_iter"] = int(self.inputs.var_cell_geom_iter)
        v_param_dict = CastepCalculation.check_castep_input(v_param_dict, auto_fix=True)

        f_param_dict = self.inputs.parameters.get_dict()
        f_param_dict["CELL"]["fix_all_cell"] = True
        f_param_dict["CELL"].pop("cell_constraints", None)
        f_param_dict["PARAM"]["geom_method"] = str(self.inputs.fix_cell_geom_method)
        f_param_dict = CastepCalculation.check_castep_input(f_param_dict,
            auto_fix=True)

        self.ctx.inputs["parameters"] = v_param_dict

        # Save the parameters for fixed cell for later
        self.ctx.fix_cell_param = f_param_dict

        self.report("Inputs validated and processed")

    def relax(self):
        """
        Relax with varable cell
        """
        inputs = dict(self.ctx.inputs)
        inputs["parameters"] = ParameterData(
            dict=self.ctx.inputs["parameters"])

        process = CastepCalculation.process()
        calc = submit(process, **inputs)
        self.report('Launching CastepCalculation <{}>'.format(calc.pid))

        return self.to_context(calcs=append_(calc))

    def swap_param(self):

        # Swap the parameters for fixed cell optimsiation
        v_calc = self.ctx.calcs[-1]
        self.report("Varcell finished. Swapping parameters")
        reuse_name = CastepCalculation.get_restart_file_relative_path(self.ctx.inputs["parameters"], False)

        # Reuse the checkfiles
        self.ctx.inputs["parameters"] = self.ctx.fix_cell_param
        self.ctx.inputs["parameters"]["PARAM"]["reuse"] = reuse_name
        self.ctx.inputs["parent_folder"] = self.ctx.calcs[-1].out.remote_folder
        self.ctx.inputs["structure"] = v_calc.out.output_structure

        self.report("Fixed cell parameters swapped. Continue with fixed cell relaxation.")

    def process_results(self):
        """
        Process the outputs
        """
        last_calc = self.ctx.calcs[-1]
        output_dict = last_calc.get_outputs_dict()
        self.out('output_structure', output_dict.get("output_structure"))
        self.out('output_parameters', output_dict.get("output_parameters"))
        self.out('output_trajectory', output_dict.get("output_trajectory"))
        self.out('remote_folder', output_dict.get("remote_folder"))
        self.out('retrieved', output_dict.get("retrieved"))

        self.report("Workchain finished.")



"""
Commandline plugin module for aiida CASTEP plugin
"""
from __future__ import print_function
import click

from aiida.cmdline.commands import data_cmd
from aiida.cmdline.dbenv_lazyloading import load_dbenv_if_not_loaded


@data_cmd.group('castep-otfg')
def otfg_cmd():
    """Commandline interface for working with OTFGData"""
    pass


@otfg_cmd.command(name="listfamilies")
@click.option(
    '--element',
    '-e',
    multiple=True,
    help=
    "Show families contenting this element only. Can be passed multiple times")
@click.option('--with_description', '-d', is_flag=True)
def listfamilies(element, with_description):
    """List avaliable OtfgData families"""
    load_dbenv_if_not_loaded()
    from aiida.orm import QueryBuilder
    from aiida_castep.data.otfg import OTFGGROUP_TYPE
    from aiida.orm import DataFactory, Group
    UspData = DataFactory("castep.otfgdata")
    q = QueryBuilder()
    q.append(UspData, tag="otfgdata")
    if element:
        q.add_filter("otfgdata", {
            "attributes.element": {
                "or": [{
                    'in': element
                }, {
                    '==': "LIBRARY"
                }]
            }
        })
    q.append(
        Group,
        tag='group',
        group_of=UspData,
        filters={'type': OTFGGROUP_TYPE},
        project=['name', 'description'])
    q.distinct()
    if q.count() > 0:
        for res in q.dict():
            group_name = res.get("group").get("name")
            group_desc = res.get("group").get("description")
            # Count the number of pseudos in this group
            q = QueryBuilder()
            q.append(
                Group, tag='thisgroup', filters={"name": {
                    'like': group_name
                }})
            q.append(UspData, project=["id"], member_of='thisgroup')

            if with_description:
                description_string = ": {}".format(group_desc)
            else:
                description_string = ""

            click.echo("* {} [{} pseudos]{}".format(group_name, q.count(),
                                                    description_string))

    else:
        click.echo("No valid OTFG pseudopotential family found.")



"""
Test for generating castep input
"""
import os
import unittest

from aiida.common.exceptions import InputValidationError
from aiida.common.folders import SandboxFolder
from aiida.orm import DataFactory, CalculationFactory
from aiida.backends.testbase import AiidaTestCase
from aiida.orm import Code
from aiida_castep.calculations.castep import CastepTSCalculation as TSCalc
from .dbcommon import BaseDataCase, BaseCalcCase

CasCalc = CalculationFactory("castep.castep")
BSCalc = CalculationFactory("castep.bs")
TSCalc = CalculationFactory("castep.ts")
StructureData = DataFactory("structure")
ParameterData = DataFactory("parameter")
KpointsData = DataFactory("array.kpoints")


class TestCastepInputGeneration(AiidaTestCase, BaseCalcCase, BaseDataCase):
    """
    Test if the input is correctly generated
    """

    @classmethod
    def setUpClass(cls):
        super(TestCastepInputGeneration, cls).setUpClass()
        cls.calc_params = {
            "computer": cls.computer,
            "resources": {
                "num_machines": 1,
                "num_mpiprocs_per_machine": 1
            }
        }

        cls.code = Code()
        cls.code.set_remote_computer_exec((cls.computer, "/x.x"))
        cls.code.store()

    def make_H2_structure(self):

        cell = ((5., 0., 0.), (0., 5., 0.), (0., 0., 5.))
        s = StructureData(cell=cell)
        s.append_atom(position=(0., 0., 0.), symbols=["H"])
        s.append_atom(position=(1., 0., 0.), symbols=["H"])
        self.H2 = s

    def test_pre_submit_checkings(self):
        """
        Test checkup before submission
        """
        pass

    def test_using_OTFG_family(self):
        """
        Test using OTFG in the input
        """
        from .utils import get_STO_structure
        STO = get_STO_structure()
        full, missing, C9 = self.create_family()
        c = CasCalc()
        pdict = self.get_default_input()
        # pdict["CELL"].pop("block species_pot")
        p = ParameterData(dict=pdict).store()
        c.use_structure(STO)
        c.use_pseudos_from_family(full)
        c.use_pseudos_from_family(C9)
        c.use_kpoints(self.get_kpoints_mesh())
        c.use_code(self.code)

        input_dict = c.get_inputs_dict()
        # Check mixing libray with acutal entry
        self.assertEqual(input_dict["pseudo_O"].entry, "C9")

        with SandboxFolder() as f:
            p = ParameterData(dict=pdict)
            c.use_parameters(p)
            input_dict = c.get_inputs_dict()
            c._prepare_for_submission(f, input_dict)

    def test_using_OTFG_mix(self):
        """
        Test using mixed OTFG library and manual values
        """
        pass

    def test_using_UpfData(self):
        """
        Test using UpfData
        """

    def generate_test_calc(self):
        """
        Return a defined Calculation
        """

        cell = ((2., 0., 0.), (0., 2., 0.), (0., 0., 2.))

        input_params = {
            "PARAM": {
                "task": "singlepoint",
                "xc_functional": "lda",
            },
            "CELL": {
                "fix_all_cell": "true",
            }
        }

        c = CasCalc(**self.calc_params).store()
        s = StructureData(cell=cell)
        s.append_atom(position=(0., 0., 0.), symbols=["Ba"])
        s.append_atom(position=(1., 0., 0.), symbols=["Ba"])
        s.store()

        p = ParameterData(dict=input_params).store()

        k = KpointsData()
        k.set_kpoints_mesh([4, 4, 4])
        k.store()

        settings_dict = {"SPINS": [0, 0]}
        c.use_settings(ParameterData(dict=settings_dict))
        c.label = "TEST CALC"
        c.description = "Test calculation for AiiDA CASTEP plugin."
        c.use_code(self.code)
        c.use_kpoints(self.get_kpoints_mesh())
        c.use_structure(s)
        c.use_parameters(p)

        return c

    def test_inputs(self):

        cell = ((2., 0., 0.), (0., 2., 0.), (0., 0., 2.))

        input_params = {
            "PARAM": {
                "task": "singlepoint",
                "xc_functional": "lda",
            },
            "CELL": {
                "fix_all_cell": "true",
                "species_pot": ("Ba Ba_00.usp",)
            }
        }

        c = CasCalc(**self.calc_params).store()
        s = StructureData(cell=cell)
        s.append_atom(position=(0., 0., 0.), symbols=["Ba"])
        s.append_atom(position=(1., 0., 0.), symbols=["Ba"])
        s.store()

        p = ParameterData(dict=input_params).store()

        k = KpointsData()
        k.set_kpoints_mesh([4, 4, 4], offset=(0.5, 0.5, 0.5))

        settings_dict = {"SPINS": [0, 0]}
        c.use_settings(ParameterData(dict=settings_dict))
        c.label = "TEST CALC"
        c.description = ("Test calculation for AiiDA CASTEP plugin. "
                         "Testing generation of inputs.")
        inputdict = c.get_inputs_dict()

        inputdict.pop("code", None)

        with SandboxFolder() as f:
            # I use the same SandboxFolder more than once because nothing
            # should be written for these failing tests

            # Missing required input nodes
            with self.assertRaises(InputValidationError):
                c._prepare_for_submission(f, inputdict)
            c.use_parameters(p)
            inputdict = c.get_inputs_dict()
            with self.assertRaises(InputValidationError):
                c._prepare_for_submission(f, inputdict)
            c.use_structure(s)
            inputdict = c.get_inputs_dict()
            with self.assertRaises(InputValidationError):
                c._prepare_for_submission(f, inputdict)
            c.use_kpoints(k)
            inputdict = c.get_inputs_dict()
            with self.assertRaises(InputValidationError):
                c._prepare_for_submission(f, inputdict)

            # Continuation keywords present - should raise exception
            inputdict = c.get_inputs_dict()
            input_params["PARAM"]["continuation"] = "default"
            p2 = ParameterData(dict=input_params)
            inputdict["parameters"] = p2
            with self.assertRaises(InputValidationError):
                c._prepare_for_submission(f, inputdict)

            c.use_code(self.code)
            inputdict = c.get_inputs_dict()
            c._prepare_for_submission(f, inputdict)

            # Check existenc of the file
            cell = f.get_abs_path(c._SEED_NAME + ".cell", check_existence=True)
            param = f.get_abs_path(c._SEED_NAME + ".param", check_existence=True)
            print("\n" + "#" * 5 + "CONTENT OF CELL FILE: " + "#" * 5)

            # Check and print out the input files
            with open(cell) as p:
                cell_content = p.read()
                print(cell_content)
                self.assertIn("fix_all_cell", cell_content)
                self.assertIn("kpoints_mp_offset", cell_content)
                self.assertIn("kpoints_mp_grid", cell_content)

            print("\n" + "#" * 5 + "CONTENT OF PARAM FILE: " + "#" * 5)
            with open(param) as p:
                param_content = p.read()
                print(param_content)
                self.assertIn("task", param_content)

            
            k.set_kpoints_mesh((2, 2, 2), (0, 0, 0))
            c.use_kpoints(k)
            inputdict = c.get_inputs_dict()
            c._prepare_for_submission(f, inputdict)
            with open(cell) as p:
                cell_content = p.read()
                print(cell_content)
                self.assertNotIn("kpoints_mp_offset", cell_content)



    @unittest.skip("Dry run test takes a long time.")
    def test_dryrun(self):
        from subprocess import call

        try:
            call(["castep.serial", "-v"])
        except OSError:
            self.skipTest("No CASTEP excutable found")

        c = self.generate_test_calc()
        full, missing, C9 = self.create_family()
        c.use_pseudos_from_family(C9)
        # Do a dry run - check if any error message is given
        with SandboxFolder() as f:
            c._prepare_for_submission(f, c.get_inputs_dict())
            self.castep_dryrun(f, c._SEED_NAME)
            self.assertFalse(f.get_content_list("*.err"))
            self.assertTrue(f.get_content_list("*.castep"))

    def castep_dryrun(self, folder, seed):
        from subprocess import call
        import os
        seed = os.path.join(folder.abspath, seed)
        call(["castep.serial", seed, "-dryrun"], cwd=folder.abspath)


class TestRestartGeneration(AiidaTestCase, BaseCalcCase, BaseDataCase):

    @classmethod
    def setUpClass(cls, *args, **kwargs):
        super(TestRestartGeneration, cls).setUpClass(*args, **kwargs)
        cls.setup_code_castep()

    def test_restart(self):

        # Initial logic of creation
        c1 = self.setup_calculation()
        c1.store_all()
        with self.assertRaises(InputValidationError):
            c2 = c1.create_restart(ignore_state=False)
        with self.assertRaises(InputValidationError):
            c2 = c1.create_restart(ignore_state=True, reuse=True)

        rmd = self.get_remote_data("H2-geom")
        rmd.store()
        c1._set_state("RETRIEVING")
        from aiida.common.links import LinkType
        rmd.add_link_from(c1, link_type=LinkType.CREATE)

        # This simply create a copy of c1
        c2 = c1.create_restart(ignore_state=True)
        c2_inp = c2.get_inputs_dict()
        c1_inp = c1.get_inputs_dict()

        # Check all inputs of c1 are indeed present in c2
        for key in c1_inp:
            self.assertIn(key, c2_inp)
            self.assertEqual(c2_inp[key].pk, c1_inp[key].pk)

        c2 = c1.create_restart(ignore_state=True, reuse=True)
        c2_inp = c2.get_inputs_dict()
        c1_inp = c1.get_inputs_dict()

        c2_param = c2_inp[c2.get_linkname("parameters")].get_dict()['PARAM']

        reuse = c2_param.pop("reuse")
        self.assertEqual(reuse, os.path.join(c2._restart_copy_to, "{}.check".format(c2._SEED_NAME)))

        with SandboxFolder() as f:
            print(c2._prepare_for_submission(f, c2_inp))

    def test_continue_from(self):
        """
        Test the continue_from function.
        It calls the same underlying function compared with
        create_restart
        """
        c1 = self.setup_calculation()
        c1.store_all()
        c2 = BSCalc.continue_from(c1, ignore_state=True, reuse=False)

        rmd = self.get_remote_data("H2-geom")
        rmd.store()
        c1._set_state("RETRIEVING")
        from aiida.common.links import LinkType
        rmd.add_link_from(c1, link_type=LinkType.CREATE)

        c3 = BSCalc.continue_from(c1, ignore_state=True)
        c1_inp = c1.get_inputs_dict()
        c2_inp = c2.get_inputs_dict()
        c3_inp = c3.get_inputs_dict()

        # Check all inputs of c1 are indeed present in c2
        for key in c1_inp:
            self.assertIn(key, c2_inp)

        for key in c1_inp:
            self.assertIn(key, c3_inp)

        with SandboxFolder() as f:
            c3._prepare_for_submission(f, c3_inp)


class TestTSCalculation(BaseCalcCase, BaseDataCase, AiidaTestCase):


    @classmethod
    def setUpClass(cls, *args, **kwargs):
        super(TestTSCalculation, cls).setUpClass(*args, **kwargs)
        cls.setup_code_castep()

    def get_default_input(self):

        input_params = {
            "PARAM": {
                "task": "transitionstatesearch",
                "xc_functional": "lda",
            },
            "CELL": {
                "fix_all_cell": "true",
                # "species_pot": ("Ba Ba_00.usp",)
            }
        }

        return input_params

    def setup_calculation(self, param=None):
        from .utils import get_STO_structure

        STO = get_STO_structure()
        full, missing, C9 = self.create_family()
        c = TSCalc()

        if param:
            pdict = param
        else:
            pdict = self.get_default_input()

        # pdict["CELL"].pop("block species_pot")
        p = ParameterData(dict=pdict)
        c.use_structure(STO)
        c.use_product_structure(STO)
        c.use_pseudos_from_family(C9)
        c.use_kpoints(self.get_kpoints_mesh())
        c.use_code(self.code)
        c.set_computer(self.computer)
        c.set_resources({"num_machines": 1, "num_mpiprocs_per_machine": 2})
        c.use_parameters(p)

        # Check mixing libray with acutal entry
        return c


    def test_input_validation(self):
        """Test input validation"""

        c = self.setup_calculation()
        pdict = c.get_inputs_dict()[c.get_linkname('parameters')].get_dict()
        pdict['PARAM']['task'] = "singlepoint"
        c.get_inputs_dict()[c.get_linkname('parameters')].set_dict(pdict)

        with SandboxFolder() as f:
            with self.assertRaises(InputValidationError):
                inputs = c.get_inputs_dict()
                c._prepare_for_submission(f, inputs)

    def test_generation(self):
        """
        Test if the block has been generated
        """
        c = self.setup_calculation()
        with SandboxFolder() as f:
            inputs = c.get_inputs_dict()
            c._prepare_for_submission(f, inputs)
            with f.open("aiida.cell") as fcell:
                cell = fcell.read()
                self.assertIn("%BLOCK POSITIONS_ABS_PRODUCT", cell)
                self.assertIn("%ENDBLOCK POSITIONS_ABS_PRODUCT", cell)


class TestBSCalculation(BaseCalcCase, BaseDataCase, AiidaTestCase):

    @classmethod
    def setUpClass(cls, *args, **kwargs):
        super(TestBSCalculation, cls).setUpClass(*args, **kwargs)
        cls.setup_code_castep()

    def get_default_input(self):

        input_params = {
            "PARAM": {
                "task": "bandstructure",
                "xc_functional": "lda",
            },
            "CELL": {
                "fix_all_cell": "true",
                # "species_pot": ("Ba Ba_00.usp",)
            }
        }

        return input_params

    def get_bs_kpoints(self):
        kpoints = KpointsData()
        kpoints.set_kpoints([[0, 0, 0], [0.5, 0.5, 0.5]])
        return kpoints

    def setup_calculation(self, param=None):
        from .utils import get_STO_structure

        STO = get_STO_structure()
        full, missing, C9 = self.create_family()
        c = BSCalc()

        if param:
            pdict = param
        else:
            pdict = self.get_default_input()

        # pdict["CELL"].pop("block species_pot")
        p = ParameterData(dict=pdict)
        c.use_structure(STO)
        c.use_pseudos_from_family(C9)
        c.use_kpoints(self.get_kpoints_mesh())
        c.use_bs_kpoints(self.get_bs_kpoints())
        c.use_code(self.code)
        c.set_computer(self.computer)
        c.set_resources({"num_machines": 1, "num_mpiprocs_per_machine": 2})
        c.use_parameters(p)

        # Check mixing libray with acutal entry
        return c

    def test_input_validation(self):
        """Test input validation"""

        c = self.setup_calculation()
        pdict = c.get_inputs_dict()[c.get_linkname('parameters')].get_dict()
        pdict['PARAM']['task'] = "singlepoint"
        c.get_inputs_dict()[c.get_linkname('parameters')].set_dict(pdict)

        with SandboxFolder() as f:
            with self.assertRaises(InputValidationError):
                inputs = c.get_inputs_dict()
                c._prepare_for_submission(f, inputs)

        c = self.setup_calculation()
        c.CHECK_EXTRA_KPN = True

        with SandboxFolder() as f:
            with self.assertRaises(InputValidationError):
                inputs = c.get_inputs_dict()
                inputs.pop("bs_kpoints")
                c._prepare_for_submission(f, inputs)

    def test_restart(self):
        """
        Test creating an restart of the calculation.
        The restart should retain the extra kpoints inputs.
        """
        c = self.setup_calculation()
        new = c.create_restart(ignore_state=True)
        self.assertEqual(type(new), type(c))
        inpname = "{}_kpoints".format(c.kpn_name)
        self.assertIn(new.get_linkname(inpname),
                      new.get_inputs_dict())

        # Extra kpn is not mandatory
        extra_kpn = c.get_inputs_dict()["bs_kpoints"]
        c._remove_link_from("bs_kpoints")
        self.assertNotIn(extra_kpn, c.get_inputs())
        # This should still succeed
        new = c.create_restart(ignore_state=True)
        self.assertNotIn(extra_kpn,
                         new.get_inputs())

    def test_bs_kpoints(self):

        c = self.setup_calculation()
        inputs = c.get_inputs_dict()
        with SandboxFolder() as f:
            c._prepare_for_submission(f, inputs)
            with f.open("aiida.cell") as cell:
                content = cell.read()
        self.assertIn("%BLOCK BS_KPOINTS_LIST", content)

    def test_bs_kpoints_mp(self):

        c = self.setup_calculation()
        inputs = c.get_inputs_dict()
        mp = KpointsData()
        mp.set_kpoints_mesh([2, 2, 2], (0.1, 0.1, 0.1))
        inputs['bs_kpoints'] = mp

        with SandboxFolder() as f:
            c._prepare_for_submission(f, inputs)
            with f.open("aiida.cell") as cell:
                content = cell.read()
        self.assertIn("{:<20}: 2 2 2".format("bs_kpoints_mp_grid"), content)
        self.assertIn("{:<20}: 0.1 0.1 0.1".format("bs_kpoints_mp_offset"), content)


class TestPot1dCalculation(BaseCalcCase, BaseDataCase, AiidaTestCase):

    def test_load_plugin(self):
        Pot1D = CalculationFactory("castep.pot1d")



"""
This module stores a single version for calculation-parser.

A single version number is ued here so whenever any change i
assed to pasers or calculation modules, the version number should
be increased.This avoids falsely matching nodes via hashes.

CHNAGELOG
0.2.3 -> FIX a typo psedu_pots -> pseudo_pots
0.2.4 -> Sort the bands and parsed kpoints from .bands file using the index given
"""

calc_parser_version = "0.2.4"



"""
Testing the parsers
"""

import os

from aiida.orm import DataFactory
from aiida.parsers import ParserFactory
from aiida.backends.testbase import AiidaTestCase
from .dbcommon import BaseCalcCase, BaseDataCase


FolderData = DataFactory("folder")


class TestCastepParser(AiidaTestCase, BaseCalcCase, BaseDataCase):

    def setUp(self):
        self.clean_db()
        self.insert_data()

    def get_data_abs_path(self):
        test_moudule = os.path.split(__file__)[0]
        data_folder = os.path.join(test_moudule, "data")
        return data_folder

    def get_dummy_outputs(self):

        retrieved = {}
        for folder in ["H2-geom", "O2-geom-spin", "Si-geom-stress", "N2-md"]:
            folderdata = FolderData()
            folderdata.replace_with_folder(
                os.path.join(self.get_data_abs_path(), folder))
            retrieved[folder] = dict(retrieved=folderdata)
        return retrieved

    def test_version_consistency(self):
        """
        Test that moudule versions are consistent
        """

        from aiida_castep.parsers.raw_parser import __version__ as prv
        from aiida_castep.parsers.castep import __version__ as pcv
        from aiida_castep.calculations.castep import __version__ as ccv
        from aiida_castep.calculations.base import __version__ as cbv
        self.assertTrue(prv == pcv == ccv == cbv)

    def test_parser_retrieved(self):
        from .utils import get_x2_structure
        self.setup_code_castep()
        calc = self.setup_calculation()

        parser = ParserFactory("castep.castep")(calc)
        retrived_folders = self.get_dummy_outputs()
        common_keys = ["cells", "positions", "forces", "symbols", "geom_total_energy"]
        md_keys = ["hamilt_energy", "kinetic_energy",
                   "velocities", "temperatures", "times"]
        geom_keys = ["geom_enthalpy"]

        for name, r in retrived_folders.items():
            if 'O2' in name:
                xtemp = 'O'
            elif 'Si' in name:
                xtemp = 'Si'
            elif 'N2' in name:
                xtemp = 'N'
            elif 'H2' in name:
                xtemp = 'H'
            # Swap the correct structure to allow desort to work
            calc.use_structure(get_x2_structure(xtemp))

            success, out = parser.parse_with_retrieved(r)
            out = dict(out)

            out_structure = out[parser.get_linkname_outstructure()]
            out_param_dict = out[parser.get_linkname_outparams()].get_dict()
            out_traj = out[parser.get_linkname_outtrajectory()]
            self.assertIn("total_energy", out_param_dict)
            self.assertIn("unit_energy", out_param_dict)
            self.assertEqual(out_param_dict["unit_energy"], "eV")
            self.assertEqual(calc.inp.structure.label, out_structure.label)
            # Check the length of sites are consistent
            self.assertEqual(len(out_structure.sites), len(out_traj.get_symbols()))

            for k in common_keys:
                self.assertIn(k, out_traj.get_arraynames())

            if name == "O2-geom-spin" or name == "Si-geom-stress":
                self.assertIn(parser.get_linkname_outbands(), out)
                bands = out[parser.get_linkname_outbands()]

                # Check if spins are handled correctly
                self.assertIn(bands.get_attr('nspins'), [1, 2])
                if bands.get_attr('nspins') == 1:
                    self.assertEqual(bands.get_attr('nkpts'), len(bands.get_bands()))
                elif bands.get_attr('nspins') == 2:
                    self.assertEqual(bands.get_attr('nkpts'), len(bands.get_bands()[0]))

                for k in geom_keys:
                    self.assertIn(k, out_traj.get_arraynames())

            if name == "N2-md":
                for k in md_keys:
                    self.assertIn(k, out_traj.get_arraynames())


class TestPot1DParser(AiidaTestCase):

    def test_load_plugin(self):
        Pot1dParser = ParserFactory("castep.pot1d")



"""
Collection of fixtures for basic setup
"""
import pytest

Ti_otfg = "Ti 3|1.8|9|10|11|30U:40:31:32(qc=5.5)"
Sr_otfg = "Sr 3|2.0|5|6|7|40U:50:41:42"
O_otfg = "O 2|1.1|15|18|20|20:21(qc=7)"


@pytest.fixture(scope="module")
def otfgdata():
    from aiida.orm import DataFactory
    return DataFactory("castep.otfgdata")


@pytest.fixture(scope="module")
def otfg():
    import aiida_castep.data.otfg as otfg
    return otfg


@pytest.fixture(scope="module")
def imps(aiida_profile):

    class Imports:

        def __init__(self):
            from aiida.orm import CalculationFactory
            from aiida.orm import DataFactory
            import aiida_castep.data.otfg as otfg
            ParameterData = DataFactory("parameter")
            for k, v in locals().items():
                setattr(self, k, v)

    return Imports()



@pytest.fixture
def localhost(aiida_profile, tmpdir):
    """Fixture for a local computer called localhost"""
    # Check whether Aiida uses the new backend interface to create collections.
    from aiida.utils.fixtures import _GLOBAL_FIXTURE_MANAGER
    from aiida.common import exceptions
    from aiida.orm import Computer
    aiida_profile = _GLOBAL_FIXTURE_MANAGER
    ldir = str(tmpdir)
    try:
        computer = Computer.get("localhost")

    except exceptions.NotExistent:
        computer = Computer()
        computer.set_name("localhost")
        computer.set_description("localhost")
        computer.set_workdir(ldir)
        computer.set_hostname("localhost")
        computer.set_scheduler_type("direct")
        computer.set_transport_type("local")
        computer.store()
    return computer


@pytest.fixture()
def code_echo(localhost):
    """Fixture of a code that just echos"""
    from aiida.orm import Code
    code = Code()
    code.set_remote_computer_exec(
        (localhost, "/bin/echo"))
    code.set_input_plugin_name("castep.castep")
    code.store()
    return code


@pytest.fixture()
def remotedata(localhost, tmpdir):
    """Create an remote data"""
    from aiida.orm.data.remote import RemoteData

    rmd = RemoteData()
    rmd.set_computer(localhost)
    rmd.set_remote_path(str(tmpdir))
    return rmd


@pytest.fixture
def kpoints_data(aiida_profile):
    """
    Return a factory for kpoints
    """
    from aiida.orm import DataFactory
    return DataFactory("array.kpoints")()


@pytest.fixture
def kpoints_mesh(kpoints_data):
    """Factory for kpoints with mesh"""
    def _kpoints_mesh(mesh, *args, **kwargs):
        kpoints_data.set_kpoints_mesh(mesh)
        return kpoints_data
    return _kpoints_mesh


@pytest.fixture
def kpoints_list(kpoints_data):
    """Factory for kpoints with mesh"""
    def _kpoints_list(klist, *args, **kwargs):
        kpoints_data.set_kpoints(klist, *args, **kwargs)
        return kpoints_data
    return _kpoints_list


@pytest.fixture
def OTFG_family_factory(aiida_profile):
    """Return a factory for upload OTFGS"""
    from aiida_castep.data.otfg import upload_otfg_family

    def _factory(otfg_entries, name, desc="TEST", **kwargs):
        upload_otfg_family(otfg_entries, name, desc, **kwargs)
        return

    return _factory


@pytest.fixture
def STO_calculation(aiida_profile, STO_structure,
                    OTFG_family_factory,
                    code_echo, imps,
                    localhost, kpoints_mesh):


    c = imps.CalculationFactory("castep.castep")()
    pdict = {"PARAM": {
        "task": "singlepoint"
    },
             "CELL": {
                 "symmetry_generate": True
             }}
    # pdict["CELL"].pop("block species_pot")
    param = imps.ParameterData(dict=pdict)
    c.use_structure(STO_structure)
    OTFG_family_factory(["C9"], "C9", stop_if_existing=False)
    c.use_pseudos_from_family("C9")
    c.use_kpoints(kpoints_mesh((3, 3, 3)))
    c.use_code(code_echo)
    c.set_computer(localhost)
    c.set_resources({"num_machines": 1, "num_mpiprocs_per_machine": 2})
    c.use_parameters(param)

    return c


def test_sto_calc(STO_calculation):
    STO_calculation.store_all()
    assert STO_calculation.pk

@pytest.fixture
def STO_structure(aiida_profile, imps):
    """Return a STO structure"""
    StructureData = imps.DataFactory("structure")
    a = 3.905

    cell = ((a, 0., 0.), (0., a, 0.), (0., 0., a))
    s = StructureData(cell=cell)
    s.append_atom(position=(0., 0., 0.), symbols=["Sr"])
    s.append_atom(position=(a / 2, a / 2, a / 2), symbols=["Ti"])
    s.append_atom(position=(a / 2, a / 2, 0.), symbols=["O"])
    s.append_atom(position=(a / 2, 0., a / 2), symbols=["O"])
    s.append_atom(position=(0., a / 2, a / 2), symbols=["O"])
    s.label = "STO"
    return s


def test_localhost_fixture(localhost):
    """
    Test the localhost fixture
    """
    localhost.name == "localhost"
    assert localhost.pk is not None


def test_code_fixture(code_echo):
    """
    Test the localhost fixture
    """
    assert code_echo.pk is not None
    code_echo.get_remote_exec_path()


def test_remotedata_fixture(remotedata):
    assert remotedata.get_remote_path()



"""
Utility functions
"""
import os

def get_STO_structure():
    """Return a STO structure"""
    from aiida.orm import DataFactory
    StructureData = DataFactory('structure')

    a = 3.905

    cell = ((a, 0., 0.), (0., a, 0.), (0., 0., a))
    s = StructureData(cell=cell)
    s.append_atom(position=(0., 0., 0.), symbols=["Sr"])
    s.append_atom(position=(a / 2, a / 2, a / 2), symbols=["Ti"])
    s.append_atom(position=(a / 2, a / 2, 0.), symbols=["O"])
    s.append_atom(position=(a / 2, 0., a / 2), symbols=["O"])
    s.append_atom(position=(0., a / 2, a / 2), symbols=["O"])
    s.label = "STO"
    return s


def get_x2_structure(x):
    """Return a O2 molecule in a box"""
    from aiida.orm import DataFactory
    StructureData = DataFactory('structure')
    a = 10

    cell = ((a, 0., 0.), (0., a, 0.), (0., 0., a))
    s = StructureData(cell=cell)
    s.append_atom(position=(0., 0., 0.), symbols=[x])
    s.append_atom(position=(1.4, 0., 0.), symbols=[x])
    s.label = "O2"
    return s


def get_data_abs_path():
    """Get the data folder path for the backend module"""
    test_moudule = os.path.split(os.path.abspath(__file__))[0]
    data_folder = os.path.join(test_moudule, "data")
    return data_folder



"""
Storing OTFG configuration as Data nodes
"""
import re
from aiida.orm import Data
from aiida.common.utils import classproperty
from aiida.common.exceptions import ValidationError

OTFGGROUP_TYPE = "data.castep.otfg.family"



def upload_otfg_family(entries,
                       group_name,
                       group_description,
                       stop_if_existing=True):
    """
    Set a family for the OTFG pseudo potential strings
    """
    from aiida.orm import Group
    from aiida.common.exceptions import UniquenessError, NotExistent
    from aiida.orm.querybuilder import QueryBuilder
    from aiida.common import aiidalogger

    # Try to retrieve a group if it exists
    try:
        group = Group.get(name=group_name, type_string=OTFGGROUP_TYPE)
        group_created = False
    except NotExistent:
        group = Group(
            name=group_name,
            type_string=OTFGGROUP_TYPE,
            )
        group_created = True

    group.description = group_description

    otfg_and_created = []
    nentries = len(entries)

    for s in entries:
        # Add it if it is just one existing data
        if isinstance(s, OTFGData):
            element, setting = s.element, s.string
        else:
            element, setting = split_otfg_entry(s)

        qb = QueryBuilder()
        qb.append(
            OTFGData,
            filters={
                'attributes.otfg_string': {
                    "==": setting
                },
                'attributes.element': {
                    "==": element
                }
            })
        existing_otfg = qb.first()

        if existing_otfg is None:

            if isinstance(s, OTFGData):
                otfg_and_created.append((s, True))
            else:
                otfg, created = OTFGData.get_or_create(
                    s, use_first=True, store_otfg=False)
                otfg_and_created.append((otfg, created))

        else:
            if stop_if_existing:
                raise ValidationError(
                    "A OTFG group cannot be added when stop_if_existing is True"
                )
            existing_otfg = existing_otfg[0]
            otfg_and_created.append((existing_otfg, False))

    # Check for unique for the complete group
    elements = [(i[0].element, i[0].string) for i in otfg_and_created]

    # Add other entries for the list to check
    if not group_created:
        for aiida_n in group.nodes:
            if not isinstance(aiida_n, OTFGData):
                continue
            elements.append((aiida_n.element, aiida_n.string))

    # Discard duplicated pairs
    elements = set(elements)
    elements_names = [e[0] for e in elements]

    # Check the uniqueness of the complete group
    if not len(elements_names) == len(set(elements_names)):
        duplicates = set(
            [x for x in elements_names if elements_names.count(x) > 1])
        dup_string = ", ".join(duplicates)
        raise UniquenessError("More than one OTFG found for the elements: " +
                              dup_string + ".")

    # If we survive here uniqueness is fine

    # Save the group
    if group_created:
        group.store()

    # Save the OTFG in the database if necessary and add them to the group

    for otfg, created in otfg_and_created:
        if created:
            otfg.store()
            aiidalogger.debug("New node {} created for OTFG string {}".format(
                otfg.uuid, otfg.string))
        else:
            aiidalogger.debug("Reusing node {} for OTFG string {}".format(
                otfg.uuid, otfg.string))

    group.add_nodes(otfg for otfg, _ in otfg_and_created)

    nuploaded = len([_ for _, created in otfg_and_created if created])

    return nentries, nuploaded


class OTFGData(Data):
    """
    Class representing an OTFG configuration

    attributes:
    string: string to be put into the cell file
    element: element that this setting is for - may not exist if we are dealing with a library.
    """

    @classmethod
    def get_or_create(cls, otfg_entry, use_first=False, store_otfg=True):
        """
        Create or retrieve OTFG from database
        :param otfg_entry: CASTEP styled OTFG entry.
        Can either be the name of library (e.g C9) or the full specification with element like:
        "O 2|1.1|15|18|20|20:21(qc=7)"

        The created OTFGData node will by default labelled by the fully entry.
        """

        in_db = cls.from_entry(otfg_entry)

        element, setting = split_otfg_entry(otfg_entry)
        if len(in_db) == 0:
            if store_otfg:
                instance = cls(string=setting, element=element).store()
            else:
                instance = cls(string=setting, element=element)
            # Automatically set the label
            instance.label = otfg_entry
            return (instance, True)

        else:
            if len(in_db) > 1:
                if use_first:
                    return (in_db[0], False)
                else:
                    pks = ", ".join([str(i.pk) for i in in_db])
                    raise ValueError(
                        "More than one duplicated OTFG data has been found. pks={}"
                        .format(pks))
            else:
                return (in_db[0], False)

    def set_string(self, otfg_string):
        """Set the full string of OTFGData instance"""
        if self.element is None:
            self.set_element("LIBRARY")
        self._set_attr("otfg_string", str(otfg_string))

    def set_element(self, element):
        """Set the element of OTFGData instance"""
        self._set_attr("element", str(element))

    def store(self, *args, **kwargs):
        self._validate()
        return super(OTFGData, self).store(*args, **kwargs)

    @property
    def string(self):
        return self.get_attr('otfg_string', None)

    @property
    def element(self):
        """Element of the OTFG. May not be available"""
        return self.get_attr('element', None)

    @property
    def entry(self):
        """Plain format of the OTFG"""
        string = self.string
        element = self.element
        if element is None or element == "LIBRARY":
            return string

        else:
            return element + " " + string

    @classmethod
    def from_entry(cls, entry):
        """
        Return a list of OTFG that matches with the string
        """

        from aiida.orm.querybuilder import QueryBuilder

        element, string = split_otfg_entry(entry)
        qb = QueryBuilder()
        qb.append(
            cls,
            filters={
                'attributes.otfg_string': {
                    '==': string
                },
                'attributes.element': {
                    '==': element
                }
            })

        return [i[0] for i in qb.all()]

    def _validate(self):
        """Validate the format of OTFG configuration"""
        super(OTFGData, self)._validate()
        if self.element is None:
            raise ValidationError("The value of element is not set. "
                                  "Set it to 'LIBRARY' manually to indicate. "
                                  "This is a library")

    @classproperty
    def otfg_family_type_string(cls):
        return OTFGGROUP_TYPE

    @classmethod
    def get_otfg_group(cls, group_name):
        """
        Return the OTFGData group with the given name.
        """
        from aiida.orm import Group

        return Group.get(
            name=group_name, type_string=cls.otfg_family_type_string)

    @classmethod
    def get_otfg_groups(cls, filter_elements=None, user=None):
        """
        Return all names of groups of type otfg, possibly with some filters.

        :param filter_elements: A string or a list of strings.
               If present, returns only the groups that contains one Usp for
               every element present in the list. Default=None, meaning that
               all families are returned.
        :param user: if None (default), return the groups for all users.
               If defined, it should be either a DbUser instance, or a string
               for the user name (that is, the user email).
        """

        from aiida.orm import Group

        group_query_params = {"type_string": cls.otfg_family_type_string}

        if user is not None:
            group_query_params['user'] = user

        if isinstance(filter_elements, basestring):
            filter_elements = [filter_elements]

        if filter_elements is not None:
            actual_filter_elements = {_.capitalize() for _ in filter_elements}
            # LIBRARY is a wild card
            actual_filter_elements.add("LBIRARY")

            group_query_params['node_attributes'] = {
                'element': actual_filter_elements
            }

        all_usp_groups = Group.query(**group_query_params)

        groups = [(g.name, g) for g in all_usp_groups]
        # Sort by name
        groups.sort()
        # Return the groups, without name
        return [_[1] for _ in groups]


def split_otfg_entry(otfg):
    """
    Split an entry of otfg in the form of element_settings
    :returns (element, entry):
    """
    otfg = otfg.strip()
    try:
        element, setting = re.split("[ _]+", otfg, 1)
    # Incase I pass a library e.g C9
    except ValueError:
        element = "LIBRARY"
        setting = otfg

    return element, setting





import os

from aiida.common.exceptions import InputValidationError
from aiida.common.folders import SandboxFolder
from aiida.orm import DataFactory, CalculationFactory
from aiida.orm import Code
from aiida.common.exceptions import MultipleObjectsError

import aiida_castep.data.otfg as otf
import aiida_castep.data.usp as usp

from .utils import get_data_abs_path

CasCalc = CalculationFactory("castep.castep")
StructureData = DataFactory("structure")
ParameterData = DataFactory("parameter")
KpointsData = DataFactory("array.kpoints")

Ti_otfg = "Ti 3|1.8|9|10|11|30U:40:31:32(qc=5.5)"
Sr_otfg = "Sr 3|2.0|5|6|7|40U:50:41:42"
O_otfg = "O 2|1.1|15|18|20|20:21(qc=7)"

otfg = DataFactory("castep.otfgdata")


class BaseDataCase(object):
    """Base to include some useful things"""

    def create_family(self):
        """Creat families for testsing"""
        Ti, Sr, O, C9 = self.get_otfgs()
        otf.upload_otfg_family([Ti.entry, Sr.entry, O.entry], "STO_FULL", "TEST", False)
        otf.upload_otfg_family([Ti.entry, Sr.entry], "STO_O_missing", "TEST", False)

        # Missing O but that's OK we have a C9 wild card here
        otf.upload_otfg_family([Ti.entry, Sr.entry, "C9"], "STO_O_C9", "TEST", False)

        return "STO_FULL", "STO_O_missing", "STO_O_C9"

    def get_otfgs(self):

        Ti, _ = otfg.get_or_create(Ti_otfg, store_otfg=False)
        Sr, _ = otfg.get_or_create(Sr_otfg, store_otfg=False)
        O, _ = otfg.get_or_create(O_otfg, store_otfg=False)
        C9, _ = otfg.get_or_create("C9", store_otfg=False)
        return Ti, Sr, O, C9


class BaseCalcCase(object):

    def get_default_input(self):

        input_params = {
            "PARAM": {
                "task": "singlepoint",
                "xc_functional": "lda",
            },
            "CELL": {
                "fix_all_cell": "true",
                # "species_pot": ("Ba Ba_00.usp",)
            }
        }

        return input_params

    def get_kpoints_mesh(self, mesh=(4, 4, 4)):

        k = KpointsData()
        k.set_kpoints_mesh(mesh)
        k.store()
        return k

    def setup_calculation(self):
        from .utils import get_STO_structure

        STO = get_STO_structure()

        full, missing, C9 = self.create_family()
        c = CasCalc()
        pdict = self.get_default_input()
        # pdict["CELL"].pop("block species_pot")
        p = ParameterData(dict=pdict).store()
        c.use_structure(STO)
        c.use_pseudos_from_family(full)
        c.use_pseudos_from_family(C9)
        c.use_kpoints(self.get_kpoints_mesh())
        c.use_code(self.code)
        c.set_computer(self.computer)
        c.set_resources({"num_machines": 1, "num_mpiprocs_per_machine": 2})
        c.use_parameters(p)

        # Check mixing libray with acutal entry
        return c

    @classmethod
    def setup_code_castep(cls):
        code = Code()
        code.set_remote_computer_exec(
            (cls.computer, "/home/bonan/appdir/CASTEP-17.2/bin/linux_x86_64_gfortran5.0--mpi/castep.mpi"))
        code.set_input_plugin_name("castep.castep")
        code.store()
        cls.code = code

    def get_remote_data(self, rel_path):
        RemoteData = DataFactory("remote")
        rmd = RemoteData()
        rmd.set_computer(self.computer)
        rmd.set_remote_path(os.path.join(get_data_abs_path(), rel_path))
        return rmd



"""
For pytest
initialise a text database and profile
"""
from __future__ import absolute_import
import tempfile
import shutil
import pytest
import os

from aiida.utils.fixtures import fixture_manager


def get_backend_str():
    """ Return database backend string.

    Reads from 'TEST_AIIDA_BACKEND' environment variable.
    Defaults to django backend.
    """
    from aiida.backends.profile import BACKEND_DJANGO, BACKEND_SQLA
    backend_env = os.environ.get('TEST_AIIDA_BACKEND')
    if not backend_env:
        return BACKEND_DJANGO
    elif  backend_env in (BACKEND_DJANGO, BACKEND_SQLA):
        return backend_env

    raise ValueError("Unknown backend '{}' read from TEST_AIIDA_BACKEND environment variable".format(backend_env))


@pytest.fixture(scope='session')
def aiida_profile():
    """setup a test profile for the duration of the tests"""
    with fixture_manager() as fixture_mgr:
        yield fixture_mgr


@pytest.fixture(scope='function')
def new_database(aiida_profile):
    """clear the database after each test"""
    yield
    aiida_profile.reset_db()


@pytest.fixture(scope='function')
def new_workdir():
    """get a new temporary folder to use as the computer's wrkdir"""
    dirpath = tempfile.mkdtemp()
    yield dirpath
    shutil.rmtree(dirpath)



"""
Module for parsing .castep file
"""

import re
import numpy as np
from aiida_castep.parsers.utils import CASTEPOutputParsingError

import logging
logger = logging.getLogger("aiida")

from .._version import calc_parser_version
__version__ = calc_parser_version

# TODO update CODATA for castep 16.1 and above

# CODATA1986 (included herein for the sake of completeness)
# taken from
#    http://physics.nist.gov/cuu/Archive/1986RMP.pdf
units_CODATA1986 = {
    'hbar': 6.5821220E-16,      # eVs
    'Eh': 27.2113961,           # eV
    'kB': 8.617385E-5,          # eV/K
    'a0': 0.529177249,          # A
    'c': 299792458,             # m/s
    'e': 1.60217733E-19,        # C
    'me': 5.485799110E-4}       # u

# CODATA2002: default in CASTEP 5.01
# (-> check in more recent CASTEP in case of numerical discrepancies?!)
# taken from
#    http://physics.nist.gov/cuu/Document/all_2002.pdf

units_CODATA2002 = {
    'hbar': 6.58211915E-16,     # eVs
    'Eh': 27.2113845,           # eV
    'kB': 8.617343E-5,          # eV/K
    'a0': 0.5291772108,         # A
    'c': 299792458,             # m/s
    'e': 1.60217653E-19,        # C
    'me': 5.4857990945E-4}      # u

units_CODATA2010 = {
    'c': 299792458,
    'hbar': 6.58211928e-16,  # eV
    'e': 1.602176565e-19,    # C
    'Av': 6.02214129e23,     # Avogadro constant
    'R': 8.3144621,          # Gas constrant
    'Eh': 27.21138505,       # eV
    'a0': 0.52917721092,     # A - bohr radius
    'me': 5.4857990946e-4,    # u
    'kB': 8.6173324E-5          # eV/K
}

# (common) derived entries
for d in (units_CODATA1986, units_CODATA2002, units_CODATA2010):
    d['t0'] = d['hbar'] / d['Eh']     # s
    d['Pascal'] = d['e'] * 1E30       # Pa
units = units_CODATA2010

unit_suffix = "_units"

SCF_FAILURE_MESSAGE = "SCF cycles failed to converge"
GEOM_FAILURE_MESSAGE = "Maximum geometry optimization cycle has been reached"
END_NOT_FOUND_MESSAGE = "CASTEP run did not reach the end of execution."
INSUFFICENT_TIME_MESSAGE = "CASTEP run terminated due to insufficient run time left."
STOP_REQUESTED_MESSAGE = "CASTEP run terminated due to requested STOP in param file."


def parse_raw_ouput(outfile, input_dict,
                    parser_opts=None, md_geom_file=None,
                    bands_file=None):
    """
    Parse an dot_castep file
    :param outfile: path to .castep file
    :param md_geom_file: path to the .geom file or .md file
    :return: A list of:

     * out_dict: a dictionary with parsed data.
     * trajectory_data: dictionary of trajectory data.
     * structure data: dictionary of cell, positions and symbols.
     * successful: a boolean that is False in case of failed calculations.

    2 different keys to check in out_dict: *parser_warning* and *warnings*.
    """

    parser_version = __version__
    parser_info = {}
    parser_info["parser_warnings"] = []
    parser_info["parser_info"] = "AiiDA CASTEP basic Parser v{}".format(
        parser_version)
    parser_info["warnings"] = []

    job_successful = True

    try:
        with open(outfile, "r") as f:
            # Store ouput lines in list
            # OK if the file is not huge
            out_lines = f.readlines()
    except IOError:
        raise CASTEPOutputParsingError(
            "Failed to open file: {}".format(outfile))

    if not out_lines:  # The file is empty
        job_successful = False

    finished_run = False

    # Use Total time as a mark for completed run
    for i, line in enumerate(out_lines[-20:]):
        # Check only the last 20 lines
        # Otherwise unfinished restarts may be seen as finished
        if "Total time" in line:
            finished_run = True
            break

    # Warn if the run is not finished
    if not finished_run:
        warning = END_NOT_FOUND_MESSAGE
        parser_info["warnings"].append(warning)
        job_successful = False

    # Parse the data and store
    try:
        out_data, trajectory_data, critical_messages = parse_castep_text_output(
            out_lines, input_dict)

        # Use data from geom file if avaliable.
        if md_geom_file is not None:
            with open(md_geom_file) as gfile:
                glines = gfile.readlines()
            geom_data = parse_geom_text_output(glines, None)
            # For geom file the second energy is the enthalpy while
            # for MD it is the approx hamiltonian (etotal + ek)
            if "geom" in md_geom_file:
                geom_data["geom_enthalpy"] = geom_data["hamilt_energy"]
            trajectory_data.update(geom_data)


        if bands_file is not None:
            bands_data = parse_dot_bands(bands_file)
        else:
            bands_data = None

    except CASTEPOutputParsingError as e:
        if not finished_run:
            parser_info["parser_warnings"].append(
                "Error while parsing the output file")
            out_data = {}
            trajectory_data = {}
            critical_messages = []

        else:  # Run finished but I have still have an error here
            raise CASTEPOutputParsingError(
                "Error while parsing ouput. Exception message: {}".format(e.message))

    # Check if any critical messages has been passed
    # If there is any cricial message we should make the run marked as FAILED
    for w in out_data["warnings"]:
        if w in critical_messages:
            job_successful = False
            break

    # Construct a structure data from the last frame
    try:
        last_cell = trajectory_data["cells"][-1]
        last_positions = trajectory_data["positions"][-1]
        symbols = trajectory_data["symbols"]

    except (KeyError, IndexError):
        # Cannot find the last geometry data
        structure_data = {}
    else:
        structure_data = dict(
            cell=last_cell, positions=last_positions, symbols=symbols)

    # Parameter data to be returned
    parameter_data = dict(out_data.items() + parser_info.items())

    # Combine the warnings
    all_warnings = out_data["warnings"] + parser_info["warnings"]

    # Make the warnings set-like e.g we don't want to repeat messages
    # Save a bit of the storage space
    all_warnings = list(set(all_warnings))
    parameter_data['warnings'] = all_warnings

    # Todo Validation of ouput data
    return [parameter_data,
            trajectory_data, structure_data,
            bands_data, job_successful]


# Re for getting unit
unit_re = re.compile("^ output\s+(\w+) unit *: *([^\s]+)")
time_re = re.compile("^(\w+) time += +([0-9.]+) s$")
parallel_re = re.compile(
    "^Overall parallel efficiency rating: [\w ]+ \(([0-9]+)%\)")
version_re = re.compile("CASTEP version ([0-9.]+)")


def parse_castep_text_output(out_lines, input_dict):
    """
    Parse ouput of .castep
    :param out_lines: a list of lines from readlines function

    :param input_dict: Control some variables. Currently support

    'n_warning_lines'- number of the lines to include for a general warning.

    :return: A list of parsed_data, trajectory_data and critical_messages:

     * parsed_data: dictionary with key values, referring to the last
       occurring quantities

     * trajectory_data: key, values of the intermediate scf steps,
       such as during geometryoptimization

     * critical_messages: a list with critical messages.

    If any is found in parsed_data["warnings"] the calculation should be considered as failed.
    """
    from collections import defaultdict

    pseudo_files = {}
    parsed_data = {}
    parsed_data["warnings"] = []

    if not input_dict:
        input_dict = {}

    # Some parameters can be controlled
    n_warning_lines = input_dict.get("n_warning_lines", 10)

    # Initialise storage space for trajectory
    # Not all is needed. But we parse as much as we can here
    trajectory_data = defaultdict(list)

    # In the format {<keywords in line>:<message to pass>}
    critical_warnings = {
        "SCF cycles performed but system has not reached the groundstate": SCF_FAILURE_MESSAGE,
        "NOSTART": "Can not find start of the calculation.",
        "STOP keyword detected in parameter file. Stop execution.": STOP_REQUESTED_MESSAGE,
        }

    # Warnings that won't result in a calculation in FAILED state
    minor_warnings = {"Warning": None,
                     "Geometry optimization failed to converge": GEOM_FAILURE_MESSAGE,
                     "Insufficient time for another iteration": INSUFFICENT_TIME_MESSAGE}


    # A dictionary witch keys we should check at each line
    all_warnings = dict(critical_warnings.items() + minor_warnings.items())
    # Create a list of keys, the order is important here
    # specific warnings can be matched first

    all_warnings_keys = list(critical_warnings.keys()) + \
        list(minor_warnings.keys())

    def pair_split(lines, spliter):
        """Split the line in to two. Ignore space around spliter"""
        lines.strip().split(spliter)

    # Split header and body
    body_start = None
    version = None
    for i, line in enumerate(out_lines):

        # Find the castep version
        if version is None:
            vmatch = version_re.search(line)
            if vmatch:
                version = vmatch.group(1)
                parsed_data["castep_version"] = version

        # Finding the units we used
        unit_match = unit_re.match(line)
        if unit_match:
            uname = unit_match.group(1)
            uvalue = unit_match.group(2)
            parsed_data["unit_" + uname] = uvalue

        if "Files used for pseudopotentials" in line:
            for line in out_lines[i + 1:]:
                if "---" in line:
                    break
                else:
                    try:
                        specie, pp_file = line.strip().split()
                    except ValueError:
                        break
                    else:
                        pseudo_files.update({specie: pp_file})
        if "Total number of ions" in line:
            parsed_data["num_ions"] = int(line.strip().split("=")[1].strip())
            continue

        if "Point group of crystal" in line:
            parsed_data["point_group"] = line.strip().split("=")[1].strip()
            continue

        if "Space group of crystal" in line:
            parsed_data["space_group"] = line.strip().split("=")[1].strip()
            continue

        if "Cell constraints" in line:
            parsed_data["cell_constraints"] = line.strip().split(":")[
                1].strip()
            continue

        if "Number of kpoints used" in line:
            parsed_data["n_kpoints"] = line.strip().split("=")[1].strip()
            continue

        if "MEMORY AND SCRATCH DISK ESTIMATES" in line:
            body_start = i
            break

    # If we don't find a start of body then there is something wrong
    if body_start is None:
        parsed_data["warnings"].append(critical_warnings["NOSTART"])
        return parsed_data, {}, critical_warnings.values()

    parsed_data.update(pseudo_pots=pseudo_files)

    def append_value_and_unit(line, name):
        """
        Extract data from = <value> <unit> line
        """
        elem = line.strip().split()
        value = float(elem[-2])
        trajectory_data[name].append(value)

    # Parse repeating information
    skip = 0
    body_lines = out_lines[body_start:]

    iter_parser = get_iter_parser()
    for count, line in enumerate(body_lines):

        # Allow sking certain number of lines
        if skip > 0:
            skip -= 1
            continue

        res_tmp = iter_parser.parse(line)
        if res_tmp:
            name, value = res_tmp[:2]
            trajectory_data[name].append(value)
            continue

        if "Calculation parallelised over" in line:
            num_cores = int(line.strip().split()[-2])
            parsed_data["parallel_procs"] = num_cores
            continue

        if "Stress Tensor" in line:
            i, stress, pressure = parse_stress_box(
                body_lines[count:count + 20])
            assert len(stress) == 3
            if "Symmetrised" in line:
                prefix = "symm_"
            else:
                prefix = ""
            trajectory_data[prefix + "pressure"].append(pressure)
            trajectory_data[prefix + "stress"].append(stress)
            skip = i

        if "Forces *******" in line:
            num_lines = parsed_data["num_ions"] + 10
            box = body_lines[count: (count + num_lines)]
            i, forces = parse_force_box(box)

            if "Constrained" in line:
                forc_name = "cons_forces"
            else:
                forc_name = "forces"
            if not forces:
                logger.error("Cannot parse force lines {}".format(box))
            trajectory_data[forc_name].append(forces)
            skip = i
            continue

        if any(i in line for i in all_warnings):
            message = [all_warnings[k] for k in all_warnings_keys
                       if k in line][0]
            if message is None:
                # CASTEP often have multiline warnings
                # Add extra lines for detail
                message = body_lines[count:count + n_warning_lines]
                message = "\n".join(message)
            parsed_data["warnings"].append(message)

        time_line = time_re.match(line)

        # Save information about time usage
        if time_line:
            time_name = time_line.group(1).lower() + "_time"
            parsed_data[time_name] = float(time_line.group(2))
            continue

        para_line = parallel_re.match(line)

        if para_line:
            parsed_data["parallel_efficiency"] = int(para_line.group(1))

    #### END OF LINE BY LINE PARSING ITERATION ####

    # remove unrelated units
    units_to_delete = []
    for key in parsed_data:
        if "unit_" in key:
            unit_for = key.split("_", 1)[1]
            delete = True
            # Check the thing this unit refers do exists
            for i in trajectory_data:
                if i == key:
                    continue
                if unit_for in i:
                    delete = False

            for i in parsed_data:
                if i == key:
                    continue
                if unit_for in i:
                    delete = False

            if delete is True:
                units_to_delete.append(key)

    for key in units_to_delete:
        parsed_data.pop(key)

    # set geom convergence state
    if GEOM_FAILURE_MESSAGE in parsed_data["warnings"]:
        parsed_data["geom_unconverged"] = True
    else:
        parsed_data["geom_unconverged"] = None

    return parsed_data, trajectory_data, critical_warnings.values()


class LineParser(object):
    """
    Parser for a line
    """

    def __init__(self, conditions):
        """initialize the Parser by passing the conditions"""
        self._cond = conditions

    def parse(self, line):
        """
        Return parsing results

        :returns: Result of the first matched Matcher object or None if no match is found
        """
        out = None
        for c in self._cond:
            out, match = c.match_pattern(line)
            if out:
                break
        return out


class Matcher(object):
    """
    Class of the condition to match the line
    """

    def __init__(self, regex, name, convfunc=None):
        """
        Initialize a Matcher object.

        :param string regex: Pattern to be matched
        :param string name: Name of the results
        """
        self.regex = re.compile(regex)
        self.convfunc = convfunc
        self.name = name


    def match_pattern(self, line):
        """
        Match pattern

        :returns: (out, match) Out is a dicationary of {self.name: <matched_value>}.
          and match is a re.MatchObject or None
        """

        match =  self.regex.match(line)
        if match:
            value = match.group(1)
            if self.convfunc:
                value = self.convfunc(value)
            out = (self.name, value)
        else:
            out = None
        return out, match

class UnitMatcher(Matcher):
    """
    The pattern of a UnitMatcher should have two groups with second group
    being the unit. The first group will be converted to float
    """

    def match_pattern(self, line):
        """
        Match the pattern
        """
        out, match = super(UnitMatcher, self).match_pattern(line)
        if out:
            unit = match.group(2)
            conv = self.convfunc if self.convfunc else float
            out = (out[0],conv(out[1]), unit)
        return out, match

def get_iter_parser():
    """
    Generate a LineParser object to parse repeating outputs
    """
    tail1 = r' *= *([0-9.+-eE]+) +(\w+)'
    mfree = UnitMatcher(r'^Final free energy \(E-TS\)' + tail1, "free_energy")
    mtotal = UnitMatcher(r'^Final energy, E' + tail1, "total_energy")
    mtotal2 = UnitMatcher(r'^Final energy' + tail1, "total_energy")
    mzeroK = UnitMatcher(r'^NB est. 0K energy \(E-0.5TS\)' + tail1, "zero_K_energy")
    spin = UnitMatcher(r'^Integrated Spin Density' + tail1, "spin_density")
    absspin = UnitMatcher(r'^Integrated \|Spin Density\|' + tail1, "abs_spin_density")
    enthalpy = UnitMatcher(r'^ *\w+: finished iteration +\d+ +with enthalpy' + tail1, "enthalpy")
    parser = LineParser([mfree, mtotal, mtotal2, mzeroK, spin, absspin, enthalpy])
    return parser

def parse_geom_text_output(out_lines, input_dict):
    """
    Parse output of .geom file

    :param out_lines: a list of lines from the readline function.
    :param dict input_dict: not in use at the moment.

    :return: key, value of the trajectories of cell, atoms, force etc
    """
    txt = out_lines
    Hartree = units['Eh'] # eV
    Bohr = units['a0'] # A
    kB = units['kB'] # eV/K
    hBar = units['hbar'] # in eV
    eV = units["e"]  # in J


    cell_list = []
    species_list = []
    geom_list = []
    forces_list = []
    energy_list = []
    hamilt_list = []
    kinetic_list = []
    pressure_list = []
    temperature_list = []
    velocity_list = []
    time_list = []

    # For a specific image
    current_pos = []
    current_species = []
    current_forces = []
    current_velocity = []
    current_cell = []
    in_header = False
    for i, line in enumerate(txt):
        if "begin header" in line.lower():
            in_header = True
            continue
        if "end header" in line.lower():
            in_header = False
            continue
        if in_header:
            continue  # Skip header lines

        sline = line.split()
        if len(sline) == 1:
            try:
                time_list.append(float(sline[0]))
            except ValueError:
                continue
        elif '<-- E' in line:
            energy_list.append(float(sline[0])) # Total energy
            hamilt_list.append(float(sline[1])) # Hamitonian (MD)
            # Kinetic energy is not blank in GEOM OPT runs
            if len(sline) == 5:
                kinetic_list.append(float(sline[2])) # Kinetic (MD)
            continue
        elif '<-- h' in line:
            current_cell.append(list(map(float, sline[:3])))
            continue
        elif '<-- R' in line:
            current_pos.append(list(map(float, sline[2:5])))
            current_species.append(sline[0])
        elif '<-- F' in line:
            current_forces.append(list(map(float, sline[2:5])))
        elif '<-- V' in line:
            current_velocity.append(list(map(float, sline[2:5])))
        elif '<-- T' in line:
            temperature_list.append(float(sline[0]))
        elif '<-- P' in line:
            pressure_list.append(float(sline[0]))
        elif not line.strip() and current_cell:
            cell_list.append(current_cell)
            species_list.append(current_species)
            geom_list.append(current_pos)
            forces_list.append(current_forces)
            current_cell = []
            current_species = []
            current_pos = []
            current_forces = []
            if current_velocity:
                velocity_list.append(current_velocity)
                current_velocity = []

    if len(species_list) == 0:
        raise CASTEPOutputParsingError("No data found in geom file")

    out =  dict(cells=np.array(cell_list) * Bohr,
                positions=np.array(geom_list) * Bohr,
                forces=np.array(forces_list) * Hartree / Bohr,
                geom_total_energy=np.array(energy_list) * Hartree,
                symbols=species_list[0],
                )

    # optional lists
    unit_V = Hartree * Bohr / hBar
    unit_K = Hartree / kB  # temperature in K
    unit_P = Hartree / (Bohr * 1e-10) ** 3 * eV
    unit_s = hBar / Hartree
    opt = {"velocities": (velocity_list, unit_V),
           "temperatures": (temperature_list, unit_K),
           "pressures": (pressure_list, unit_P),
           "hamilt_energy": (hamilt_list, Hartree),
           "times": (time_list, unit_s),
           "kinetic_energy": (kinetic_list, Hartree)
           }
    for key, value in opt.items():
        if value[0]:
            out.update({key: np.array(value[0]) * value[1]})
    return out


force_match = re.compile(
    "^ +\* +(\w+) +([0-9]+) +([0-9.\-+]+) +([0-9.\-+]+) +([0-9.\-+]+) +\*")
stress_match = re.compile(
    "^ +\* +(\w+) +([0-9.\-+]+) +([0-9.\-+]+) +([0-9.\-+]+) +\*")


def parse_force_box(lines):
    """
    Parse a box of the forces
    :param lines: a list of upcoming lines

    :return: A list of number of lines read and the forces
    """

    forces = []
    for i, line in enumerate(lines):

        if "Forces" in line:
            continue

        if "***********************" in line:
            break

        match = force_match.match(line)
        if match:
            forces.append([float(match.group(i)) for i in range(3, 6)])

    return i, forces


def parse_stress_box(lines):
    """
    Parse a box of the stress
    :param lines: a list of upcoming lines

    :return: a list of  [number of lines read, stress_tensor, pressure]
    """

    stress = []
    pressure = None
    for i, line in enumerate(lines):
        if "Stress" in line:
            continue

        if "Cartisian components" in line:
            unit = line.strip().split()[-2]
            unit = unit[1:-2]
            continue

        if "**********************" in line:
            break

        match = stress_match.match(line)
        if  match:
            stress.append([float(match.group(i)) for i in range(2, 5)])
        elif "Pressure" in line:
            lsplit = line.strip().split()
            pressure = float(lsplit[-2])

    return i, stress, pressure


def parse_dot_bands(file_path):
    """
    Parse an CASTEP bands file
    Extract Kpoints and each bands for each kpoints.
    This is a generic parsing function. Return python builtin types.

    :param str file_path: Path of the file to parse
    :return: A list of bands_info, kpoints and bands:

     * bands_info: A dictionary for information of bands.

     * kpoints: A list of kpoints. In the format [kpoint index, coordinats x 3 kpoint weight]

     * bands: A list of bands. Each band has a list of actual
       eigenvalues for each spin components. E.g nkpoints, nspins, neigns.

    Note that the atomic units are used in the bands file
    """
    fh = open(file_path)

    i_finish = None
    cell = []
    bands_info = {}
    for i, line in enumerate(fh):
        if not line.strip():
            continue
        if "Number of k-points" in line:
            nkps = line.strip().split()[-1]
            bands_info['nkpts'] = int(float(nkps))
            continue
        if "Number of spin components" in line:
            nspin = line.strip().split()[-1]
            bands_info['nspins'] = int(float(nspin))
            continue
        if "Number of electrons" in line:
            nelec = line.strip().split()[-1]
            bands_info['nelecs'] = float(nelec)
            continue
        if "Number of eigenvalues" in line:
            neigns = line.strip().split()[-1]
            bands_info['neigns'] = int(float(neigns))
            continue
        if "Fermi energ" in line:  # TODO support for spin polarisation
            efermi = line.strip().split()[-1]
            bands_info['efermi'] = float(efermi)
            continue
        if "Unit cell" in line:
            i_finish = i + 3
            continue

        # Added the cell
        if i_finish:
            cell.append([float(n) for n in line.strip().split()])
        if i == i_finish:
            break
    bands_info['cell'] = cell

    # Now parser the body
    kpoints = []
    bands = []
    this_band = []
    this_spin = []
    for line in fh:
        if "K-point" in line:
            # We are not at the first kpoints
            if kpoints:
                # Save the result from previous block
                this_band.append(this_spin)
                bands.append(this_band)
                this_spin = []
                this_band = []
            kline = line.strip().split()
            kpoints.append([float(n) for n in kline[1:]])
            continue
        if "Spin component" in line:
            # Check if we are at the second spin
            if this_spin:
                this_band.append(this_spin)
                this_spin = []
            continue

        ls = line.strip()
        if not ls:
            continue
        this_spin.append(float(ls))

    fh.close()

    # Save the last set of results
    this_band.append(this_spin)
    bands.append(this_band)

    # Do some sanity checks
    assert int(nkps) == len(kpoints), "Missing kpoints"
    assert len(bands) == len(kpoints), "Missing bands for certain kpoints"

    for n, b in enumerate(bands):
        assert len(b) == int(
            nspin), "Missing spins for kpoint {}".format(n + 1)
        for i, s in enumerate(b):
            assert len(s) == int(neigns), ("Missing eigenvalues "
                                           "for kpoint {} spin {}".format(n + 1, i + 1))

    return bands_info, kpoints, bands



"""
A module of base class for CASTEP calculations
"""
from __future__ import print_function

import copy
import os
import time
from textwrap import TextWrapper

from aiida.backends.utils import get_current_profile
from aiida.common.datastructures import CalcInfo, CodeInfo
from aiida.common.exceptions import InputValidationError, MultipleObjectsError
from aiida.common.utils import classproperty
from aiida.orm import DataFactory
from aiida_castep.calculations.datastructure import CellFile, ParamFile
from aiida_castep.calculations.helper import CastepHelper
from aiida_castep.data import OTFGData, UspData, get_pseudos_from_structure

from .utils import get_castep_ion_line, _lowercase_dict, _uppercase_dict

from .._version import calc_parser_version
__version__ = calc_parser_version

StructureData = DataFactory("structure")
ParameterData = DataFactory("parameter")
KpointsData = DataFactory("array.kpoints")
UpfData = DataFactory("upf")
RemoteData = DataFactory("remote")


class BaseCastepInputGenerator(object):
    """
    Baseclass for generating CASTEP inputs
    """
    # This may not apply to CASTEP runs
    _PSEUDO_SUBFOLDER = "./pseudo/"
    # Castep output is in the
    _PARENT_CALC_SUBFOLDER = "./parent/"
    _PREFIX = 'aiida'
    _INPUT_FILE_NAME = "aiida.cell"
    _OUTPUT_FILE_NAME = "aiida.castep"
    _SEED_NAME = 'aiida'

    # Additional files that should always be retrieved if they exist
    _internal_retrieve_list = [
        "*.err", "*.den_fmt", "*-out.cell", "*.pdos_bin",
    ]

    # in restarts, will not copy but use symlinks
    _default_symlink_usage = True

    # in restarts, it will copy from the parent the following
    _restart_copy_from = "./"

    # in restarts, it will copy the previous folder in the following one
    _restart_copy_to = _PARENT_CALC_SUBFOLDER

    # Default verbosity; change in subclasses
    _default_verbosity = 1

    # whether we are automatically validating the input parameters
    _auto_input_validation = True

    retrieve_dict = {
        "phonon": [".phonon"],
        "magres": [".magres"],
        "transitionstatesearch": [".ts"],
        "molecular dynamics": [".md"],
        "moleculardynamics": [".md"],
        "geometryoptimisation": [".geom"],
        "geometryoptimization": [".geom"],
        "spectral": [".ome_bin", ".dome_bin"],
    }

    @classproperty
    def _baseclass_use_methods(cls):
        """
        This will be manually added to the _use_methods in each subclass
        """
        return {
            "structure": {
                'valid_types': StructureData,
                'additional_parameter': None,
                'linkname': 'structure',
                'docstring': "Choose the input structure to use",
            },
            "settings": {
                'valid_types': ParameterData,
                'additional_parameter': None,
                'linkname': 'settings',
                'docstring': "Use an additional node for special settings",
            },
            "parameters": {
                'valid_types':
                ParameterData,
                'additional_parameter': None,
                'linkname': 'parameters',
                'docstring':
                ("Use a node that specifies the input parameters.",
                 "Usually needs to contain PARAM and CELL keys.")
            },
            "parent_folder": {
                'valid_types':
                RemoteData,
                'additional_parameter':
                None,
                'linkname':
                'parent_calc_folder',
                'docstring': ("Use a remote folder as parent folder (for "
                              "restarts and similar"),
            },
            "pseudo": {
                'valid_types': (UpfData, OTFGData, UspData),
                'additional_parameter':
                "kind",
                'linkname':
                cls._get_linkname_pseudo,
                'docstring':
                ("Use a node for the  pseudopotential of one of "
                 "the elements in the structure. You have to pass "
                 "an additional parameter ('kind') specifying the "
                 "name of the structure kind (i.e., the name of "
                 "the species) for which you want to use this "
                 "pseudo. You can pass either a string, or a "
                 "list of strings if more than one kind uses the "
                 "same pseudo"),
            },
        }

    def _generate_header_lines(self, other_nodes=None):
        """
        Generate header lines to go into param and cell files
        """

        wrapper = TextWrapper(initial_indent="# ", subsequent_indent="# ")
        time_str = time.strftime("%H:%M:%S %d/%m/%Y %Z")
        lines = [
            "##### Generated by aiida_castep {} #####".format(time_str),
            "#         author: Bonan Zhu (bz240@cam.ac.uk)",
            "# "
            "# AiiDA User: {}".format(self.get_user().get_full_name()),
            "# AiiDA profile: {}".format(get_current_profile()),
            "# Information of the calculation node",
            "# type: {}".format(self._query_type_string[:-1]),
            "# pk: {}".format(self.pk),
            "# uuid: {}".format(self.uuid),
            "# label: {}".format(self.label),
            "# description:",
        ]

        # Fix for SQLA backend - default of node.description is None
        if self.description:
            lines.extend(wrapper.wrap(self.description))
        lines.append("")

        # additional information of the input nodes
        if other_nodes:
            lines.append("## Information of input nodes used:")
        for node in other_nodes:
            if node:
                node_lines = [
                    "# ", "# type: {}".format(node._query_type_string[:-1]),
                    "# pk: {}".format(node.pk), "# uuid: {}".format(node.uuid),
                    "# label: {}".format(node.label), "# description:"
                ]
                _desc = node.description
                if _desc:
                    node_lines.extend(wrapper.wrap(_desc))
                node_lines.append("")
                lines.extend(node_lines)

        lines.append("# END OF HEADER")

        return lines

    def _generate_CASTEPinputdata(self,
                                  parameters,
                                  structure,
                                  pseudos,
                                  settings_dict,
                                  settings=None,
                                  kpoints=None,
                                  **kwargs):
        """
        This method creates the content of an input file in the
        CASTEP format
        Generated input here should be generic to all castep calculations

        :param parameters, ParameterData: Input goes to the .cell or .param file
        :param settings_dict: A dictionary of the settings used for generation
        :param pseudos: A dictionary of pseudo potential Data for each kind
        :param structure: A StructureData instance
        :param kpoints: A KpointsData node, optional
        """
        local_copy_list_to_append = []

        # The input dictionary should be {"CELL": {"fix_all_cell": True},
        # "PARAM":{"opt_strategy" : "speed"}}

        # The following lines enforces the case of names
        # EACH entry is written as in lower case

        input_params = _uppercase_dict(
            parameters.get_dict(), dict_name="parameters")
        input_params = {
            k: _lowercase_dict(v, dict_name=k)
            for k, v in input_params.iteritems()
        }

        # Check if there are keywords that need to be blocked

        # Set verbosity to 1.
        # Parser may not work if verbosity is not 1
        input_params["PARAM"]["iprint"] = input_params["PARAM"].get(
            "iprint", self._default_verbosity)

        # Set run_time using define value for this calcualtion
        run_time = self.get_max_wallclock_seconds()
        if run_time:
            n_seconds = run_time * 0.95
            n_seconds = (
                n_seconds // 60) * 60  # Round down to the nearest minutes
            # Do not do any thing if calculated time is less than 1 hour
            if n_seconds < 3600:
                pass
            elif "run_time" not in input_params["PARAM"]:
                input_params["PARAM"]["run_time"] = int(n_seconds)

        # Set the default comment using the label of this calculation
        comment_str = self.label
        if "comment" not in input_params["PARAM"]:
            input_params["PARAM"]["comment"] = comment_str

        # ========= Start to prepare input site data ======

        cellfile = CellFile()
        paramfile = ParamFile()
        # --------- CELL ----------
        cell_vector_list = []
        for vector in structure.cell:
            cell_vector_list.append(("{0:18.10f} {1:18.10f} "
                                     "{2:18.10f}".format(*vector)))

        cellfile["LATTICE_CART"] = cell_vector_list

        # --------- ATOMIC POSITIONS---------
        # for kind in structure.kinds:
        atomic_position_list = []
        mixture_count = 0
        # deal with initial spins
        spin_list = settings_dict.pop("SPINS", None)
        label_list = settings_dict.pop("LABELS", None)

        for i, site in enumerate(structure.sites):
            # get  the kind of the site
            kind = structure.get_kind(site.kind_name)

            # Position is always needed
            pos = site.position
            try:
                name = kind.symbol
            # If we are dealing with mixed atoms
            except ValueError:
                name = kind.symbols
                mixture_count += 1

            if spin_list:
                spin = spin_list[i]
            else:
                spin = None

            # deal with labels
            if label_list:
                label = label_list[i]
            else:
                label = None
            # Get the line of positions_abs block
            line = get_castep_ion_line(
                name,
                pos,
                label=label,
                spin=spin,
                occupation=kind.weights,
                mix_num=mixture_count)

            # Append the line to the list
            atomic_position_list.append(line)

        # End of the atomic position block
        cellfile["POSITIONS_ABS"] = atomic_position_list

        # Check the consistency of spin in parameters
        if spin_list:
            total_spin = sum(s for s in spin_list if s)
            param_spin = input_params["PARAM"].get("spin", None)
            if param_spin is not None:
                # If spin is specified - check consistency
                if param_spin != total_spin:
                    raise InputValidationError(
                        "Inconsistent spin in cell and param files."
                        "Total spin: {} in cell file but {} in param file".
                        format(total_spin, param_spin))
            else:
                # If no spin specified, do it automatically
                # Note that we don't check if spin polarized calculation is
                # requested in the first place
                input_params["PARAM"]["spin"] = total_spin

        # --------- KPOINTS ---------
        if self._use_kpoints:
            try:
                mesh, offset = kpoints.get_kpoints_mesh()
                has_mesh = True

            except AttributeError:
                try:
                    kpoints_list = kpoints.get_kpoints()
                    num_kpoints = len(kpoints_list)
                    has_mesh = False
                    if num_kpoints == 0:
                        raise InputValidationError(
                            "At least one k points must be provided")
                except AttributeError:
                    raise InputValidationError(
                        "No valid kpoints have been found")

                try:
                    _, weights = kpoints.get_kpoints(also_weights=True)

                except AttributeError:
                    import numpy as np
                    weights = np.ones(num_kpoints, dtype=float) / num_kpoints

            kpoints_line_list = []
            if has_mesh is True:
                cellfile["kpoints_mp_grid"] = "{} {} {}".format(*mesh)
                if offset != [0., 0., 0.]:
                    cellfile["kpoints_mp_offset"] = "{} {} {}".format(*offset)
            else:
                for kpoint, weight in zip(kpoints_list, weights):
                    kpoints_line_list.append("{:18.10f} {:18.10f} "
                                             "{:18.10f} {:18.10f}".format(
                                                 kpoint[0], kpoint[1],
                                                 kpoint[2], weight))
                cellfile["KPOINTS_LIST"] = kpoints_line_list

        # --------- PSEUDOPOTENTIALS --------
        # Check if we are using UPF pseudos
        # Now only support simple elemental pseudopotentials
        if pseudos:
            symbols = set()  # All of the symbols
            species_pot_list = []
            for kind in structure.kinds:
                for s in kind.symbols:
                    symbols.add(s)

            # Make symbols unique
            for s in symbols:
                ps = pseudos[s]  # Get the pseudopotential object

                # If we are dealing with a UpfData object
                if isinstance(ps, (UpfData, UspData)):
                    species_pot_list.append("{:5} {}".format(s, ps.filename))
                    # Add to the copy list
                    local_copy_list_to_append.append((ps.get_file_abs_path(),
                                                      ps.filename))

                # If we are using OTFG, just add the string property of it
                if isinstance(ps, OTFGData):
                    species_pot_list.append("{:5} {}".format(s, ps.string))

            cellfile["SPECIES_POT"] = species_pot_list

        # --------- PARAMETERS in cell file---------
        for key, value in input_params["CELL"].iteritems():

            if "species_pot" in key:
                if pseudos:
                    raise MultipleObjectsError(
                        "Both species_pot and pseudos are provided")
                self.logger.warning(
                    "Pseudopotentials directly defined in CELL dictionary")

            # Constructing block keywrods
            # We identify the key should be treated as a block it is not a string and has len() > 0
            cellfile[key] = value

        # Parameters for PARAM files
        paramfile.update(input_params["PARAM"])
        param_header = self._generate_header_lines([parameters])
        paramfile.header = param_header

        # Added header to .cell file
        cell_file_header = self._generate_header_lines(
            [parameters, structure, kpoints, settings])
        cellfile.header = cell_file_header

        return cellfile, paramfile, local_copy_list_to_append

    def _prepare_for_submission(self, tempfolder, inputdict):
        """
        Routine to be called when create the input files and other stuff

        :param tempfolder: a aiida.common.folders.Folder subclass where
                           the plugin should put all its files.
        :param inputdict: a dictionary with the input nodes, as they would
                be returned by get_inputs_dict (without the Code!)
        """

        local_copy_list = []

        # TODO implement remote copy for restart calculations
        remote_copy_list = []
        remote_symlink_list = []

        try:
            parameters = inputdict.pop(self.get_linkname('parameters'))
            # Validate the parameters
            if self._auto_input_validation is True:
                self.check_castep_input(parameters.get_dict(), auto_fix=False)

        except KeyError:
            raise InputValidationError(
                "No parameters specified for this calculation")
        if not isinstance(parameters, ParameterData):
            raise InputValidationError(
                "parameters is not of type ParameterData")

        try:
            structure = inputdict.pop(self.get_linkname('structure'))
        except KeyError:
            raise InputValidationError(
                "No structure specified for this calculation")
        if not isinstance(structure, StructureData):
            raise InputValidationError(
                "structure is not of type StructureData")

        if self._use_kpoints:
            try:
                kpoints = inputdict.pop(self.get_linkname('kpoints'))
            except KeyError:
                raise InputValidationError(
                    "No kpoints specified for this calculation")
            if not isinstance(kpoints, KpointsData):
                raise InputValidationError(
                    "kpoints is not of type KpointsData")
        else:
            kpoints = None

        # Settings can be undefined, and defaults to an empty dictionary
        settings = inputdict.pop(self.get_linkname('settings'), None)
        if settings is None:
            settings_dict = {}
        else:
            if not isinstance(settings, ParameterData):
                raise InputValidationError(
                    "settings, if specified, must be of "
                    "type ParameterData")
            # Settings converted to uppercase
            settings_dict = _uppercase_dict(
                settings.get_dict(), dict_name='settings')

        # Set up pseodo potentials
        # The existance of pseudopoential is optional
        pseudos = {}
        # I create here a dictionary that associates each kind name to a pseudo
        for link in inputdict.keys():
            if link.startswith(self._get_linkname_pseudo_prefix()):
                kindstring = link[len(self._get_linkname_pseudo_prefix()):]
                kinds = kindstring.split('_')
                the_pseudo = inputdict.pop(link)
                if not isinstance(the_pseudo, (UpfData, UspData, OTFGData)):
                    raise InputValidationError(
                        "Pseudo for kind(s) {} is not of "
                        "supported ".format(",".join(kinds)))
                for kind in kinds:
                    if kind in pseudos:
                        raise InputValidationError(
                            "Pseudo for kind {} passed "
                            "more than one time".format(kind))
                    pseudos[kind] = the_pseudo

        # Check parent calc folder
        parent_calc_folder = inputdict.pop(
            self.get_linkname('parent_folder'), None)
        if parent_calc_folder is not None:
            if not isinstance(parent_calc_folder, RemoteData):
                raise InputValidationError("parent_calc_folder, if specified, "
                                           "must be of type RemoteData")

        # If requested to reuse, check if the parent_calc_folder is defined
        param_dict = parameters.get_dict()["PARAM"]
        require_parent = False
        for k in param_dict:
            if str(k).lower() in ["reuse", "continuation"]:
                require_parent = True
                break
        if parent_calc_folder is None and require_parent:
            raise InputValidationError(
                "No parent calculation folder passed"
                " for restart calculation using reuse/continuation")

        # Check if a code is specified
        try:
            code = inputdict.pop(self.get_linkname('code'))
        except KeyError:
            raise InputValidationError(
                "No code specified for this calculation")

        # Here, there should be no more parameters...
        # But in case there is, check if this is something not implemented
        # in this base class
        if inputdict:
            for key in inputdict:
                if key not in self._use_methods:
                    raise InputValidationError(
                        "The following input data nodes are "
                        "unrecognised: {}".format(inputdict.keys()))
        ##############################
        # END OF INITIAL INPUT CHECK #
        ##############################

        # Generate input file
        cellfile, paramfile, pseudo_copy_list = self._generate_CASTEPinputdata(
            parameters,
            structure,
            pseudos,
            settings_dict,
            settings=settings,
            kpoints=kpoints,
            **inputdict)

        local_copy_list.extend(pseudo_copy_list)

        cell_input_filename = tempfolder.get_abs_path(self._SEED_NAME +
                                                      ".cell")

        param_input_filename = tempfolder.get_abs_path(self._SEED_NAME +
                                                       ".param")

        with open(cell_input_filename, "w") as incell:
            incell.write(cellfile.get_string())

        with open(param_input_filename, "w") as inparam:
            inparam.write(paramfile.get_string())

        # IMPLEMENT OPERATIONS FOR RESTART
        symlink = settings_dict.pop('PARENT_FOLDER_SYMLINK',
                                    self._default_symlink_usage)

        if parent_calc_folder is not None:
            if symlink:
                remote_list = remote_symlink_list
            else:
                remote_list = remote_copy_list
            remote_list.append(
                (parent_calc_folder.get_computer().uuid,
                 os.path.join(parent_calc_folder.get_remote_path(),
                              self._restart_copy_from), self._restart_copy_to))

        calcinfo = CalcInfo()

        calcinfo.uuid = self.uuid

        # COPY/SYMLINK LISTS
        calcinfo.local_copy_list = local_copy_list
        calcinfo.remote_copy_list = remote_copy_list
        calcinfo.remote_symlink_list = remote_symlink_list

        # SET UP extra CMDLINE arguments
        cmdline_params = settings_dict.pop("CMDLINE", [])

        # Extra parameters are added after the seed for CASTEP
        calcinfo.cmdline_params = [self._SEED_NAME] + list(cmdline_params)

        # CASTEP don't have any STDOUT etc when running calculations
        # Error is shown in the *.err file

        # Construct codeinfo instance
        codeinfo = CodeInfo()
        codeinfo.cmdline_params = [self._SEED_NAME] + list(cmdline_params)
        codeinfo.code_uuid = code.uuid
        calcinfo.codes_info = [codeinfo]

        # Retrieve by default the .castep file and the bands file
        calcinfo.retrieve_list = []

        calcinfo.retrieve_list.append(self._SEED_NAME + ".castep")
        calcinfo.retrieve_list.append(self._SEED_NAME + ".bands")

        settings_retrieve_list = settings_dict.pop("ADDITIONAL_RETRIEVE_LIST",
                                                   [])
        calcinfo.retrieve_list.extend(settings_retrieve_list)

        # If we are doing geometryoptimisation retrieved the geom file and -out.cell file
        calculation_mode = parameters.get_dict().get("PARAM", {}).get(
            "task", "")

        # dictionary for task specific file retrieve
        task_extra = self.retrieve_dict.get(calculation_mode.lower(), [])
        for suffix in task_extra:
            settings_retrieve_list.append(self._SEED_NAME + suffix)

        # Retrieve output cell file if requested
        if parameters.get_dict().get("PARAM", {}).get("write_cell_structure"):
            settings_retrieve_list.append(self._SEED_NAME + "-out.cell")

        calcinfo.retrieve_list += settings_retrieve_list
        calcinfo.retrieve_list += self._internal_retrieve_list

        # Remove parser options in the setting dictionary
        # At the moment parser options are not used here
        try:
            Parserclass = self.get_paaserclass()
            parser = Parserclass(self)
            parser_opts = parser.get_parser_settings_key()
            settings_dict.pop(parser_opts)
        except (KeyError, AttributeError):
            pass

        if settings_dict:
            raise InputValidationError(
                "The following keys have been found in "
                "the settings input node, but were not understood: {}".format(
                    ",".join(settings_dict.keys())))

        return calcinfo

    def _set_parent_remotedata(self, remotedata):
        """
        Used to set a parent remotefolder
        """
        from aiida.common.exceptions import ValidationError

        if not isinstance(remotedata, RemoteData):
            raise ValueError('remotedata must be a RemoteData')

        # complain if another remotedata is already found
        input_remote = self.get_inputs(node_type=RemoteData)
        if input_remote:
            raise ValidationError("Cannot set several parent calculation to a "
                                  "{} calculation".format(
                                      self.__class__.__name__))

        self.use_parent_folder(remotedata)

    def create_restart(self,
                       ignore_state=False,
                       restart_type="restart",
                       reuse=False,
                       use_symlink=None,
                       use_output_structure=False,
                       use_castep_bin=False,
                       param_update=None,
                       param_delete=None):
        """
        A method to help restart the calculation by creating a new one.
        Return a new calculation with all the essential input nodes in the unstored stated.

        CASTEP has two modes of 'restart', activated by setting CONTINUATION or REUSE keywords in .param file:

        CONTINUATION
          Restart from the end of the last run.
          Only limited set of parameters can be modified.
          If unmodifiable parameters were changed, they are ignored.
          E.g changes of task, nextra_bands, cut_off_energy will be ignored. This is often used for geometry optimisation or md runs.

        REUSE
          Essentially making a new calculation with parameters read from .cell and .param file.
          Data from ``castep_bin`` or ``check`` will be reused to initialise the model of the new run when applicable. This is often used for bandstructure, dos, spectral calculation.

        Note both castep_bin and check file may be used.
        They are almost the same except castep_bin does not have wavefunctions stored.


        :param bool ignore_state: Ignore the state of parent calculation
        :param str restart_type: "continuation" or "restart". If set to continuation the child calculation has keyword 'continuation' set.
        :param bool reuse: Wether we want to reuse the previous calculation.
            only applies for "restart" run
        :param bool parent_folder_symlink: if True, symlink are used instead of hard copies of the files. Default given be cin._default_symlink_usage
        :param bool use_output_structure: if True, the output structure of parent calculation is used as the input of the child calculation.
            This is useful for photon/bs calculation.
        :param calc_class: Class of the new calculation, e.g can restart into a CastepBSCalculation.
        :param param_update: A dictionary of the parameter keywords to be updated.
            The key can be put into the top level and will be assigned accordingly.
        :param param_delete: A list of the keys to be removed from param file or cell file
        :param keep_input_model: If True, keeps the input model of the parent calculation.
            i.e. use input check file rather that the check file written.

        ..see also:: :py:func:`aiida_castep.calculations.base.create_restart_`
        """

        cout = create_restart_(
            self,
            ignore_state=ignore_state,
            restart_type=restart_type,
            reuse=reuse,
            use_symlink=use_symlink,
            use_output_structure=use_output_structure,
            use_castep_bin=use_castep_bin,
            param_update=param_update,
            param_delete=param_delete)
        return cout

    @classmethod
    def continue_from(cls, cin, *args, **kwargs):
        """
        Create a new calculation as a continuation from a given calculation.
        This is effectively an "restart" for CASTEP and a lot of the parameters
        can be tweaked. For example, conducting bandstructure calculation from
        finished geometry optimisations.
        CASTEP has two modes of 'restart', activated by setting CONTINUATION or REUSE keywords in .param file.

        ..see also:: :py:func:`aiida_castep.calculations.base.create_restart_`
        """
        cout = create_restart_(cin, calc_class=cls, *args, **kwargs)

        return cout

    @classmethod
    def check_castep_input(cls, input_dict, auto_fix=False):
        """
        Validate if the keywords are valid castep keywords
        Also try to convert the parameter diction in a
        "standardized" form
        """
        helper = cls.get_input_helper()
        out = helper.check_dict(input_dict, auto_fix)
        return out

    @classmethod
    def get_input_helper(cls, version=None):
        """
        Get a CASTEP help object.
        :param version: Specify a specific version of the help
        :type version: str

        :return: A CastepHelper object
        """
        helper = CastepHelper(version)
        return helper

    @classmethod
    def _get_linkname_pseudo_prefix(cls):
        """
        The prefix for the name of the link used for each pseudo before the kind name
        """
        return "pseudo_"

    @classmethod
    def _get_linkname_pseudo(cls, kind):
        """
        The name of the link used for the pseudo for kind 'kind'.
        It appends the pseudo name to the pseudo_prefix, as returned by the
        _get_linkname_pseudo_prefix() method.

        :note: if a list of strings is given, the elements are appended
          in the same order, separated by underscores

        :param kind: a string (or list of strings) for the atomic kind(s) for
            which we want to get the link name
        """
        # If it is a list of strings, and not a single string: join them
        # by underscore
        if isinstance(kind, (tuple, list)):
            suffix_string = "_".join(kind)
        elif isinstance(kind, basestring):
            suffix_string = kind
        else:
            raise TypeError("The parameter 'kind' of _get_linkname_pseudo can "
                            "only be a string or a list of strings")
        return "{}{}".format(cls._get_linkname_pseudo_prefix(), suffix_string)

    @classmethod
    def get_restart_file_relative_path(cls,
                                       param_data_dict,
                                       use_castep_bin=False):
        """
        Returns a relative path of the restart file
        """
        parent_check_name = param_data_dict["PARAM"].get("check_point", None)
        if parent_check_name is None:
            if use_castep_bin:
                parent_check_name = cls._SEED_NAME + ".castep_bin"
            else:
                parent_check_name = cls._SEED_NAME + ".check"

        return os.path.join(cls._PARENT_CALC_SUBFOLDER, parent_check_name)

    def use_pseudos_from_family(self, family_name):
        """
        Set the pseudo to use for all atomic kinds, picking pseudos from the
        family with name family_name.


        :note: The structure must already be set.

        :param family_name: the name of the group containing the pseudos
        """
        from collections import defaultdict

        try:
            structure = self._get_reference_structure()
        except AttributeError:
            raise ValueError(
                "Structure is not set yet! Therefore, the method "
                "use_pseudos_from_family cannot automatically set "
                "the pseudos")

        # A dict {kind_name: pseudo_object}
        # But we want to run with use_pseudo(pseudo, kinds)

        kind_pseudo_dict = get_pseudos_from_structure(structure, family_name)

        # Group the species by pseudo
        # pseudo_dict will just map PK->pseudo_object
        pseudo_dict = {}
        # Will contain a list of all species of the pseudo with given PK
        pseudo_species = defaultdict(list)

        for kindname, pseudo in kind_pseudo_dict.iteritems():
            pseudo_dict[pseudo.pk] = pseudo
            pseudo_species[pseudo.pk].append(kindname)

        # Finally call the use_pseudo method
        for pseudo_pk in pseudo_dict:
            pseudo = pseudo_dict[pseudo_pk]
            kinds = pseudo_species[pseudo_pk]
            # I set the pseudo for all species, sorting alphabetically
            self.use_pseudo(pseudo, sorted(kinds))

    def _get_reference_structure(self):
        """
        Used to get the reference structure to obtain which
        pseudopotentials to use from a given family using
        use_pseudos_from_family.

        :note: this method can be redefined in a given subclass
               to specify which is the reference structure to consider.
        """
        return self.get_inputs_dict()[self.get_linkname('structure')]


def create_restart_(cin,
                    ignore_state=False,
                    restart_type="restart",
                    reuse=False,
                    use_symlink=None,
                    use_output_structure=False,
                    use_castep_bin=False,
                    calc_class=None,
                    param_update=None,
                    param_delete=None,
                    keep_input_model=False):
    """
    A function to help restart the calculation by creating a new one.
    Return a new calculation with all the essential input nodes in the unstored stated.

    CASTEP has two modes of 'restart', activated by setting CONTINUATION or REUSE keywords in .param file:

    CONTINUATION
      Restart from the end of the last run.
      Only limited set of parameters can be modified.
      If unmodifiable parameters were changed, they are ignored.
      E.g changes of task, nextra_bands, cut_off_energy will be ignored. This is often used for geometry optimisation or md runs.

    REUSE
      Essentially making a new calculation with parameters read from .cell and .param file.
      Data from ``castep_bin`` or ``check`` will be reused to initialise the model of the new run when applicable. This is often used for bandstructure, dos, spectral calculation.

    Note both castep_bin and check file may be used.
    They are almost the same except castep_bin does not have wavefunctions stored.

    :param bool ignore_state: Ignore the state of parent calculation
    :param str restart_type: "continuation" or "restart". If set to continuation the child calculation has keyword 'continuation' set.
    :param bool reuse: Wether we want to reuse the previous calculation.
        only applies for "restart" run
    :param bool parent_folder_symlink: if True, symlink are used instead of hard copies of the files.
        Default given be cin._default_symlink_usage
    :param bool use_output_structure: if True, the output structure of parent calculation is used as the input of the child calculation.
        This is useful for photon/bs calculation.
    :param calc_class: Class of the new calculation. E.g can restart into a CastepBSCalculation.
    :param param_update: A dictionary of the parameter keywords to be updated.
        The key can be put into the top level and will be assigned accordingly.
    :param param_delete: A list of the keys to be removed from param file or cell file
    :param keep_input_model: If True, keeps the input model of the parent calculation.
        i.e. use input check file rather that the check file written.
    """

    from aiida.common.datastructures import calc_states

    if cin.get_state(from_attribute=True) != calc_states.FINISHED:
        if not ignore_state:
            raise InputValidationError(
                "Calculation to be restarted must be in the {} state."
                "use ignore_state keyword to override this".format(
                    calc_states.FINISHED))

    if restart_type == "continuation":
        # If we do a conitniuation we actually have to re-use the file from previous run
        reuse = True
    elif restart_type == "restart":
        pass
    else:
        raise RuntimeError("Unkown type of restart {}".format(restart_type))

    if use_symlink is None:
        use_symlink = cin._default_symlink_usage

    # Duplicate the calculation
    if calc_class is None:
        cout = cin.duplicate()
    else:
        cout = calc_class()
        cout.set_computer(cin.get_computer())

    calc_inp = cin.get_inputs_dict()  # Input nodes of parent calculation

    # Find the RemoteData of this calculation
    if reuse:
        remote_folders = [
            o for o in cin.get_outputs() if isinstance(o, RemoteData)
        ]
        if len(remote_folders) > 1:
            raise InputValidationError("More than one output RemoteData found "
                                       "in calculation {}".format(cin.pk))
        elif not remote_folders:
            raise InputValidationError("No output RemoteData found "
                                       "in calculation {}".format(cin.pk))
        else:
            remote_folder = remote_folders[0]

    # We just keep the existing link
    elif keep_input_model:
        remote_folder = cin.get_inputs_dict().get(cin.get_linkname("parent_folder"))
    else:
        remote_folder = None

    if remote_folder:
        cout._set_parent_remotedata(remote_folder)

    # Use the out_put structure if required
    if use_output_structure:
        try:
            cout.use_structure(cin.out.output_structure)
        except AttributeError:
            cout.logger.warning(
                "Warning: No output structure found. Fallback to input structure"
            )
            cout.use_structure(calc_inp[cin.get_linkname('structure')])
    else:
        cout.use_structure(calc_inp[cin.get_linkname('structure')])

    # Copy the kpoints
    # NOTE this need to be changed for restarting BS with bs_kpoints etc
    # Move to another method to allow subclass modification
    if cin._use_kpoints:
        cout.use_kpoints(calc_inp[cin.get_linkname('kpoints')])
    cout.use_code(calc_inp[cin.get_linkname('code')])

    # copy the settings dictionary
    try:
        old_settings_dict = calc_inp[cin.get_linkname('settings')].get_dict()
    except KeyError:
        old_settings_dict = {}

    # Use deep copy to ensure two dictionaries are independent
    new_settings = copy.deepcopy(old_settings_dict)

    if use_symlink != cout._default_symlink_usage:
        new_settings['PARENT_FOLDER_SYMLINK'] = use_symlink

    if new_settings:
        if new_settings != old_settings_dict:
            # Link to an new settings
            settings = ParameterData(dict=new_settings)
            cout.use_settings(settings)
        else:
            # Nothing changed, just use the old settings
            cout.use_settings(calc_inp[cin.get_linkname('settings')])

    # SETUP the keyword in PARAM file
    parent_param = calc_inp[cin.get_linkname('parameters')]

    in_param_dict = parent_param.get_dict()
    if reuse:
        if restart_type == "restart":
            # Set keyword reuse, pop any continuation keywords
            in_param_dict['PARAM'].pop('continuation', None)
            # Define the name of reuse here
            in_param_dict['PARAM'][
                "reuse"] = cin.get_restart_file_relative_path(
                    in_param_dict, use_castep_bin)
        elif restart_type == "continuation":
            # Do the opposite
            in_param_dict['PARAM'].pop('reuse', None)
            in_param_dict['PARAM'][
                'continuation'] = cin.get_restart_file_relative_path(
                    in_param_dict, use_castep_bin)
        else:
            raise RuntimeError("Invalid restart type: " + restart_type)

    elif keep_input_model:
        # No need to change the continuation/reuse field
        pass
    else:
        # In this case we simply create a identical calculation
        # But we should discard reuse / continuation
        in_param_dict['PARAM'].pop('continuation', None)
        in_param_dict['PARAM'].pop('reuse', None)

    # UPDATE the key words
    if param_update:
        from .helper import HelperCheckError
        helper = cout.get_input_helper()
        dict_update, not_found = helper._from_flat_dict(param_update)
        if not_found:
            suggest = [helper.get_suggestion(i) for i in not_found]
            error_string = "Following keys are invalid -- "
            for error_key, sug in zip(not_found, suggest):
                error_string += "{}: {}; ".format(error_key, sug)
            raise HelperCheckError(error_string)
        else:
            in_param_dict["PARAM"].update(dict_update["PARAM"])
            in_param_dict["CELL"].update(dict_update["CELL"])

    # Remove keywords
    if param_delete:
        for key in param_delete:
            tmp1 = in_param_dict["PARAM"].pop(key, None)
            tmp2 = in_param_dict["CELL"].pop(key, None)
            if (tmp1 is None) and (tmp2 is None):
                raise RuntimeError("Key {} not found".format(key))

    # If we have not changed anything, link the old dict
    if in_param_dict == parent_param.get_dict():
        cout.use_parameters(parent_param)
    else:
        cout.use_parameters(ParameterData(dict=in_param_dict))

    # Use exactly the same pseudopotential data
    for linkname, input_node in cin.get_inputs_dict().iteritems():
        if isinstance(input_node, (UpfData, UspData, OTFGData)):
            cout.add_link_from(input_node, label=linkname)

    return cout



#!/usr/bin/env python

import subprocess
import os
from os.path import abspath, dirname, join
from setuptools import setup, find_packages
import json

def git_version():
    def _minimal_ext_cmd(cmd):
        # construct minimal environment
        env = {}
        for k in ['SYSTEMROOT', 'PATH', 'HOME']:
            v = os.environ.get(k)
            if v is not None:
                env[k] = v
        # LANGUAGE is used on win32
        env['LANGUAGE'] = 'C'
        env['LANG'] = 'C'
        env['LC_ALL'] = 'C'
        out = subprocess.Popen(cmd, stdout=subprocess.PIPE, env=env).communicate()[0]
        return out

    try:
        out = _minimal_ext_cmd(['git', 'rev-parse', '--short', 'HEAD'])
        GIT_REVISION = out.strip().decode('ascii')
    except OSError:
        GIT_REVISION = "Unknown"

    return GIT_REVISION

if __name__ == '__main__':
    # Provide static information in setup.json
    # such that it can be discovered automatically
    ROOT = abspath(dirname(__file__))
    with open(join(ROOT, 'setup.json'), 'r') as info:
        kwargs = json.load(info)

    # Determine if we are install in the git repository
#    GIT_VERSION = git_version()
#    if GIT_VERSION != "Unkown":
#        kwargs["version"] = kwargs["version"] + "-" + GIT_VERSION

    # Included the README.md as the long description
    with open(join(ROOT, 'README.md'), 'r') as f:
        long_desc = f.read()
    setup(
        packages=find_packages(),
        long_description=long_desc,
        long_description_content_type='text/markdown',
        include_package_data=True,
        **kwargs
    )
